/*  This source file is (c) 2010 - 2012 Volker Schatz
    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details)              */

/*! \file

    \brief Implementation of application-level functions
    
    This file implements the initialisation and event queue functions declared
    in iliad_app.h.  Two functions are run as separate threads for running the
    GTK event loop and receiving inter-process communication (IPC) from the
    toolbar and pagebar.
*/

#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <liberdm/display.h>
#include <liberipc/eripc.h>
#include <liberipc/eripcviewer.h>
#include <liberipc/eripctoolbar.h>
#include <liberipc/eripcpagebar.h>
#include "iliad_api.h"
#include "iliad_app.h"

static GtkWidget *window= NULL;
static Display *display= NULL;
static Window xwindow;
static pthread_t gtk_main_thread, evq_ipc_thread;

#define EVQ_SIZE    1024

/*! \brief Event queue

  Common event queue for relevant GTK events and IPC events.  Contains a FIFO
  of iliad_event structs, a mutex and a condition variable for access control
  between threads.
*/
static struct {
    iliad_event q[EVQ_SIZE];
    iliad_event *write, *read;
    pthread_mutex_t ex;
    pthread_cond_t not_empty;
}
ev_q;

/*!  \brief Flags indicating receipt of synchronisation events from toolbar and
            pagebar.    */
static int sync_events;

void *gtk_main_func(void *arg);
int evq_init();
gboolean evq_gtk(GtkWidget *widget, GdkEvent *event, gpointer ptr);
void *evq_ipc(void *arg);
int evq_enqueue(const iliad_event *ev);
int evq_exit();

/*!
  \brief Initialise GTK and X and create main window

  This function creates an application main window of the correct size and
  obtains the GTK and X window handle and the X display.  Besides, the event
  mask is set, the GTK event handlers installed, the common event queue is set
  up and the GTK and IPC event loop threads are created.

  \param prgname Name or path of the command as which this application was
  started.  If this is not correct, the application will fail to receive most
  GTK events for unknown reasons.  Pass argv[0] here.

  \return 0 OK, != 0 error.  In the latter case, the application should
  terminate.
*/
int app_init(char *prgname)
{
    char *pseudo_argv[]= { prgname };
    char **pseudo_argp= pseudo_argv;
    int pseudo_argc= 1;
    GdkColor white;
    int status;

    if( !g_thread_supported() )
        g_thread_init(NULL);
    gdk_threads_init();
    gtk_init(&pseudo_argc, &pseudo_argp);
    window= gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "iliad_app");
    gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);
    gtk_window_set_resizable(GTK_WINDOW(window), FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(window), 0);
    gtk_widget_set_size_request(GTK_WIDGET(window), SCREEN_WIDTH, CLIENT_AREA);
    gdk_color_parse("white", &white);
    gtk_widget_modify_bg(GTK_WIDGET(window), GTK_STATE_NORMAL, &white);
    gtk_widget_set_events(GTK_WIDGET(window),
        GDK_EXPOSURE_MASK | GDK_BUTTON_MOTION_MASK |
        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
        GDK_KEY_PRESS_MASK );
    GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(window), GTK_DOUBLE_BUFFERED);
    gtk_widget_show(GTK_WIDGET(window));

    display= gdk_x11_display_get_xdisplay(gdk_display_get_default());
    xwindow= GDK_WINDOW_XWINDOW(window->window);

    gtk_signal_connect( GTK_OBJECT(window), "key_press_event",
                          GTK_SIGNAL_FUNC(evq_gtk), NULL);
    gtk_signal_connect( GTK_OBJECT(window), "button_press_event",
                          GTK_SIGNAL_FUNC(evq_gtk), NULL);
    gtk_signal_connect( GTK_OBJECT(window), "button_release_event",
                          GTK_SIGNAL_FUNC(evq_gtk), NULL);
    gtk_signal_connect( GTK_OBJECT(window), "expose_event",
                          GTK_SIGNAL_FUNC(evq_gtk), NULL);
    gtk_signal_connect( GTK_OBJECT(window), "delete_event",
                          GTK_SIGNAL_FUNC(evq_gtk), NULL);
/*    gtk_signal_connect( GTK_OBJECT(window), "motion_notify_event",
                          GTK_SIGNAL_FUNC(evq_gtk), NULL);    */

    status= evq_init();
    status |= pthread_create(&gtk_main_thread, NULL, gtk_main_func, NULL);
    status |= pthread_create(&evq_ipc_thread, NULL, evq_ipc, NULL);

    return !(window && display && !status);
}

/*!
  Separate thread which runs the GTK event loop by calling gtk_main().  The
  argument is ignored, and NULL is always returned.
*/
void *gtk_main_func(void *arg)
{
    gdk_threads_enter();
    gtk_main();
    gdk_threads_leave();
    return NULL;
}

/*!
  \brief Clean up before application terminates.

  The GTK and IPC event handling threads are terminated.  This function must be
  called outside any gdk_threads_enter()...gdk_threads_leave() region since it
  calls these functions itself (and the GTK lock is not re-entrant).
*/
void app_exit()
{
    pthread_cancel(evq_ipc_thread);
    gdk_threads_enter();
    gtk_main_quit();
    gdk_threads_leave();
    pthread_join(gtk_main_thread, NULL);
    evq_exit();
}

/*!
  \brief Returns the GTK handle of the application window.

  app_init() must have been called successfully before this function can be
  used.
*/
GtkWidget *app_window()
{
    return window;
}

/*!
  \brief Returns the X display.

  app_init() must have been called successfully before this function can be
  used.
*/
Display *app_xdisplay()
{
    return display;
}

/*!
  \brief Returns the X handle of the application window.

  app_init() must have been called successfully before this function can be
  used.
*/
Window app_xwindow()
{
    return xwindow;
}


/*!
  Initialise event FIFO read and write pointers and mutex and condition
  variable.

  \return 0 OK, !=0 error from pthread_mutex_init() or pthread_cond_init().
*/
int evq_init()
{
    ev_q.read= ev_q.write= ev_q.q;
    return pthread_mutex_init(&ev_q.ex, NULL) ||
        pthread_cond_init(&ev_q.not_empty, NULL);
}


/*!
  GTK event handler function.  Receives GTK events and converts them to
  corresponding iliad_event type events to be written to the queue.  Button
  press events are held back until the corresponding button release event to
  form a combined pen drag event.  Key press events are translated to the
  mnemonic constants in iliad_app.h.

  \param widget ignored
  \param event GTK event
  \param ptr ignored
  \return always TRUE
*/
gboolean evq_gtk(GtkWidget *widget, GdkEvent *event, gpointer ptr)
{
    static unsigned short dragging= 0, startx, starty;

    iliad_event ev;

/*  /usr/local/arm/oe/arm-linux/include/gtk-2.0/gdk/gdkevents.h
    /usr/local/arm/oe/arm-linux/include/gtk-2.0/gdk/gtksignal.h   */
    switch(event->type)
    {
        case GDK_DELETE:
                ev.type= EV_SYS;
                ev.sys.subtype= EV_SYS_DELETE;
            break;
        case GDK_EXPOSE:
                ev.type= EV_SYS;
                ev.sys.subtype= EV_SYS_EXPOSE;
            break;
        case GDK_KEY_PRESS:
                ev.type= EV_KEY;
                switch(event->key.keyval) {
                    case GDK_F5:ev.key.key= EV_KEY_ASCEND;
                                ev.key.long_press= 0;
                        break;
                    case GDK_Home:ev.key.key= EV_KEY_ASCEND;
                                  ev.key.long_press= 1;
                        break;
                    case GDK_Next:ev.key.key= EV_KEY_FORW;
                                  ev.key.long_press= 0;
                        break;
                    case GDK_F1:ev.key.key= EV_KEY_FORW;
                                ev.key.long_press= 1;
                        break;
                    case GDK_Prior:ev.key.key= EV_KEY_BACK;
                                   ev.key.long_press= 0;
                        break;
                    case GDK_F2:ev.key.key= EV_KEY_BACK;
                                ev.key.long_press= 1;
                        break;
                    case GDK_Up:ev.key.key= EV_KEY_UP;
                                ev.key.long_press= 0;
                        break;
                    case GDK_F4:ev.key.key= EV_KEY_UP;
                                ev.key.long_press= 1;
                        break;
                    case GDK_Return:ev.key.key= EV_KEY_DOT;
                                    ev.key.long_press= 0;
                        break;
                    case GDK_F6:ev.key.key= EV_KEY_DOT;
                                ev.key.long_press= 1;
                        break;
                    case GDK_Down:ev.key.key= EV_KEY_DOWN;
                                  ev.key.long_press= 0;
                        break;
                    case GDK_F3:ev.key.key= EV_KEY_DOWN;
                                ev.key.long_press= 1;
                        break;
                    default:return TRUE;
                    /*  Only four buttons and the page bar are reported
                    to the application, the others are handled by the
                    content lister.  */
                }
            break;
        case GDK_BUTTON_PRESS:
                dragging= 1;
                startx= (unsigned short)event->button.x;
                starty= (unsigned short)event->button.y;
            return TRUE;
        case GDK_BUTTON_RELEASE:
                if( !dragging )
                    return TRUE;
                ev.type= EV_PEN;
                ev.pen.xfrom= startx;
                ev.pen.yfrom= starty;
                ev.pen.xto= (unsigned short)event->button.x;
                ev.pen.yto= (unsigned short)event->button.y;
                dragging= 0;
            break;
        default:return TRUE;
    }
    evq_enqueue(&ev);
    return TRUE;
}


#define IPC_BUF_SIZE    1024
/*!
  IPC event queue thread.  Iliad IPC messages are received from the toolbar
  and the pagebar and are converted to iliad_event type events to be queued.

  \param arg ignored
  \return This function never returns, but is terminated by evq_exit().
*/
void *evq_ipc(void *arg)
{
    erServerChannel_t channel;
    char ipc_buf[IPC_BUF_SIZE];
    erIpcCmd_t cmd;
    iliad_event ev;
    int dummy, bufsize;

    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &dummy);
    if( erIpcOpenServerChannel(ER_PDF_VIEWER_CHANNEL, &channel) )
        return NULL;
    while( 13 )
    {
        bufsize= IPC_BUF_SIZE;
        erIpcGetMessage(channel, ipc_buf, &bufsize);
        if( vwrParseCommand(ipc_buf, &cmd) < 0 )
            continue;
        switch( cmd.cc )
        {
            case ccVwrToolbar:
                    ev.type= EV_TOOLBAR;
                    ev.toolbar.icon= (unsigned short)atoi(cmd.arg[0]);
                    ev.toolbar.state= (unsigned short)atoi(cmd.arg[1]);
                break;
            case ccVwrJumpToPage:
                    ev.type= EV_PAGEBAR;
                    ev.pagebar.page= (unsigned short)atoi(cmd.arg[1]);
                break;
            case ccVwrToolbarSync:
                    sync_events |= SYNC_TOOLBAR;
                break;
            case ccVwrPagebarSync:
                    sync_events |= SYNC_PAGEBAR;
                break;
            default:continue;
        }
        evq_enqueue(&ev);
    }
}

/*!
  This function adds an event to the common (GTK + IPC) event queue after
  obtaining access by locking the mutex.  If the queue is already full, the
  event is discarded.

  \param ev event to queue
  \return 0 on success, 1 on overflow.
*/
int evq_enqueue(const iliad_event *ev)
{
    iliad_event *nextwrite;

    nextwrite= ev_q.write+1;
    if( nextwrite - ev_q.q >= EVQ_SIZE )
        nextwrite -= EVQ_SIZE;
    if( nextwrite == ev_q.read )        /*  discard latest event on overflow  */
        return 1;
    pthread_mutex_lock(&ev_q.ex);
    *ev_q.write= *ev;
    ev_q.write= nextwrite;
    pthread_cond_signal(&ev_q.not_empty);
    pthread_mutex_unlock(&ev_q.ex);
    return 0;
}

/*!
  Clean up event queue prior to application exit by destroying mutex and
  condition variable for event queue access control.

  \return 0 OK, !=0 error from pthread_cond_destroy() or pthread_mutex_destroy()
*/
int evq_exit()
{
    return pthread_cond_destroy(&ev_q.not_empty) ||
        pthread_mutex_destroy(&ev_q.ex);
}

/*!
  \brief Returns number of events in the event queue

  If the return value is not 0, a subsequent app_wait_event() will not block.

  \return number of available events
*/
int app_events_pending()
{
    int nevents= ev_q.write - ev_q.read;
    if( nevents < 0 )
        nevents += EVQ_SIZE;
    return nevents;
}

/*!
  \brief Return next event from the event queue

  If no event is available, this function blocks until there is.

  \param ev pointer to write event struct to
*/
void app_wait_event(iliad_event *ev)
{
    pthread_mutex_lock(&ev_q.ex);
    while( ev_q.read == ev_q.write )
        pthread_cond_wait(&ev_q.not_empty, &ev_q.ex);
    *ev= *ev_q.read;
    ++ev_q.read;
    if( ev_q.read - ev_q.q >= EVQ_SIZE )
        ev_q.read -= EVQ_SIZE;
    pthread_mutex_unlock(&ev_q.ex);
}


/*!
    \brief Synchronise with toolbar and/or pagebar

    Send synchronise messages to toolbar and/or pagebar and wait for receipt of
    synchronise events from them.  Times out after approximately one second.
    This works only via the reporting of sync events by evq_ipc() and therefore
    requires app_init() to have been called before.

    Synchronising with the tool/pagebar seems to cause their latest event to be
    resent.  It is apparently not necessary for ensuring their region of the
    display is updated.

    \param mask Bit mask; generate from SYNC_TOOLBAR and SYNC_PAGEBAR

    \return SYNC_ERROR if a synchronise message could not be sent, otherwise
    SYNC_OK or SYNC_TIMEOUT.
*/
int sync_bars(int mask)
{
    erClientChannel_t channel;
    int count;

    mask &= (SYNC_TOOLBAR | SYNC_PAGEBAR);
    sync_events= 0;
    if( mask & SYNC_TOOLBAR ) {
        channel= toolbar_channel();
        if( !channel || tbSynchronise(channel, UAID, ccVwrToolbarSync) )
            return SYNC_ERROR;
    }
    if( mask & SYNC_PAGEBAR ) {
        channel= pagebar_channel();
        if( !channel || pbSynchronise(channel, UAID) )
            return SYNC_ERROR;
    }
    for( count= 0; count < 20; ++count ) {
        if( sync_events == mask )
            return SYNC_OK;
        usleep(50000);
    }
    if( sync_events == mask )
        return SYNC_OK;
    return SYNC_TIMEOUT;
}


/*!
  \brief Debug output file handle

  This function returns a writable file handle to one of the following, in
  order of declining priority: the terminal on which a current SSH session is
  running (/dev/pts/0), the file /tmp/iliad_api.debug (opened to append new
  messages), or stderr.  The files are opened on demand, and opening the first
  two is retried every time this function is called.

  \return File handle
*/
FILE *app_debug()
{
    static FILE *debug;

    if( !debug ) {
        debug= fopen("/dev/pts/0", "w");
        if( !debug )
            debug= fopen("/tmp/iliad_api.debug", "a");
    }
    return debug? debug : stderr;
}


