/*  This source file is (c) 2010 - 2012 Volker Schatz
    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details)              */

/*! \file

  \brief Iliad API functions for application initialisation and event handling.

  This application-level library contains functionality that every application
  will have to perform, such as initialising GTK, opening a full-screen window
  and receiving events from GTK and other sources.  A common event queue is
  maintained which receives events from GTK, the toolbar and the pagebar, and
  from which the application can read when it is ready.  A union event type
  declared here can contain all common kinds of events.

  Required linker options:  -lpthread -leripc -lerdm -lgtk-x11-2.0 -lgthread-2.0
*/

#ifndef ILIAD_APP_H
#define ILIAD_APP_H

#include <X11/Xlib.h>
#include <gtk/gtkstyle.h>
#include <gdk/gdkkeysyms.h>

int app_init(char *prgname);
void app_exit();
GtkWidget *app_window();
Display *app_xdisplay();
Window app_xwindow();

#define EV_SYS      0
#define EV_KEY      1
#define EV_PEN      2
#define EV_TOOLBAR  3
#define EV_PAGEBAR  4

#define EV_SYS_DELETE   0
#define EV_SYS_EXPOSE   1

/*!  \brief System event type, currently limited to exposure and delete events.  */
struct iliad_event_sys {    /*! == EV_SYS */
    unsigned short  type;   /*! Can be one of the EV_SYS_... constants  */
    unsigned short  subtype;
};

#define EV_KEY_ASCEND   0
#define EV_KEY_FORW     1
#define EV_KEY_BACK     2
#define EV_KEY_UP       3
#define EV_KEY_DOT      4
#define EV_KEY_DOWN     5

/*!  \brief Key (= Iliad button) event.  */
struct iliad_event_key {    /*! == EV_KEY */
    unsigned short  type;   /*! Can be one of the EV_KEY_... constants  */
    unsigned short  key;    /*! Flag indicating long button presses  */
    unsigned short  long_press;
};

/*! \brief Pen drag or click event  */
struct iliad_event_pen {    /*! == EV_PEN */
    unsigned short  type;
    /*! Coordinates where the pen is put down and raised  */
    unsigned short  xfrom, yfrom, xto, yto;
};

/*! \brief Toolbar IPC event  */
struct iliad_event_toolbar {    /*! == EV_TOOLBAR */
    unsigned short  type;
    /*! Icon index of type enum e_iconID_t from liberipc/eripctoolbar.h  */
    unsigned short  icon;
    /*! Icon state of type enum e_iconState_t from liberipc/eripctoolbar.h  */
    unsigned short  state;
};

/*! \brief Pagebar IPC event  */
struct iliad_event_pagebar {    /*! == EV_PAGEBAR */
    unsigned short  type;       /*! Clicked page number */
    unsigned short  page;
};

/*! \brief Union event type, subtypes distinguished by type member */
typedef union {
    unsigned short              type;
    struct iliad_event_sys      sys;
    struct iliad_event_key      key;
    struct iliad_event_pen      pen;
    struct iliad_event_toolbar  toolbar;
    struct iliad_event_pagebar  pagebar;
}
iliad_event;

int app_events_pending();
void app_wait_event(iliad_event *ev);

#define SYNC_TOOLBAR    1
#define SYNC_PAGEBAR    2

#define SYNC_OK         0
#define SYNC_ERROR      1
#define SYNC_TIMEOUT    2

int sync_bars(int mask);

FILE *app_debug();

#endif

