/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "serialreader.h"

/*!
    \brief Allocate buffer to be enlarged in the future

    Malloc()s a buffer with the size given by the size entry of its argument.

    \param sp Pointer to struct which contains the requested size and will
    receive the pointer to the buffer.
*/
void autoalloc(autospace *sp)
{
    sp->buf= malloc(sp->size);
}


/*!
    \brief Enlarge buffer

    Calls realloc() to enlarge a buffer.  The increment entry in its argument
    determines by how many bytes the buffer is extended.  If it is zero, the
    size is doubled.  If the reallocation fails, the size entry is left
    unchanged.

    WARNING: The address of the allocated space may change as a result of this
    function, rendering all pointers into it invalid.

    \param Pointer to struct containing the pointer to the old buffer, the
    current size and the increment.
*/
void autoextend(autospace *sp)
{
    void *newbuf;
    unsigned newsize;

    if( sp->increment )
        newsize= sp->size + sp->increment;
    else
        newsize= sp->size * 2;
    newbuf= realloc(sp->buf, newsize);
    if( newbuf ) {
        sp->buf= newbuf;
        sp->size= newsize;
    }
}


/*!
    \brief Append element to array after reallocating as needed

    Creates or enlarges an array as needed before appending an element.  On the
    first call, the buf pointer in the array struct must be NULL, but the
    elemsize and maxelem entries should be set to the element size and maximum
    number of elements to allocate space for.

    Subsequent calls append one element pointed to by the second argument.  If
    necessary, the array is enlarged using autoextend() first.  If the
    reallocation fails, the element cannot be appended.

    WARNING: The address of the space allocated for the array may change as a
    result of this function, rendering all pointers into it invalid.

    \param ary Pointer to array data struct containing pointer, element size,
    current and maximum allocated number of elements
    \param elem Pointer to new element to append
*/
void autoappend(autoarray *ary, const void *elem)
{
    if( !ary->buf ) {
        if( !ary->elemsize )
            ary->elemsize= sizeof(int);
        if( !ary->maxelem )
            ary->maxelem= 8;
        ary->reserved0= 0;
        ary->totalsize= ary->elemsize * ary->maxelem;
        autoalloc((autospace*)ary);
        if( ary->buf ) {
            memcpy(ary->buf, elem, ary->elemsize);
            ary->nelem= 1;
        }
        else
            ary->nelem= 0;
    }
    else {
        if( ary->nelem >= ary->maxelem ) {
            autoextend((autospace*)ary);
            ary->maxelem= ary->totalsize / ary->elemsize;
            if( ary->nelem >= ary->maxelem )    /* realloc failed */
                return;
        }
        memcpy((char*)ary->buf+ary->elemsize*ary->nelem, elem, ary->elemsize);
        ++ary->nelem;
    }
}


/*!
    \brief Debug message stream

    This functions opens a debug message output stream on demand and returns
    it.  Messages are appended to /tmp/serialreader.diag.

    \return Debug message output stream, or NULL if it could not be opened.
*/
FILE *stddiag(void)
{
    static FILE *stream;

    if( !stream ) {
        stream= fopen("/tmp/serialreader.diag", "a");
        if( stream )
            fprintf(stream, "@ %u seconds after epoch, PID %u\n",
                            (unsigned)time(NULL), getpid());
    }
    return stream;
}


/*!
    \brief Concatenate path and file name

    \param path Path name which need not include trailing slash
    \param file File or directory name

    \return Newly allocated string containing path and file concatenated with a
    slash in between.  Has to be freed by the caller.
*/
char *catpath(const char *path, const char *file)
{
    char *all;
    unsigned pathsize, size;

    pathsize= strlen(path);
    size= pathsize + 1 + strlen(file) + 1;
    all= malloc(size);
    if( !all )
        return NULL;
    strcpy(all, path);
    all[pathsize]= '/';
    strcpy(all+pathsize+1, file);
    return all;
}


/*!
    \brief Read a file into RAM

    Allocates RAM and reads all of a file into it.  A zero byte is appended
    (useful for text files), and the size is written to the second argument.
    If unsuccessful, file error messages go into stddiag().

    \param filename File name
    \param filesize Pointer to long variable for file/buffer size (or NULL)

    \return Pointer to file data, or NULL if the operation failed.
*/
void *load_file(const char *filename, long *filesize)
{
    FILE *in;
    void *buf;
    long size;

    in= fopen(filename, "r");
    if( !in ) {
        fprintf(stddiag(), "Could not open `%s' for reading.\n", filename);
        return NULL;
    }
    fseek(in, 0, SEEK_END);
    size= ftell(in);
    fseek(in, 0, SEEK_SET);
    buf= malloc(size+1);
    if( !buf ) {
        fprintf(stddiag(), "Could not allocate memory for loading `%s'.\n", filename);
        fclose(in);
        return NULL;
    }
    if( ferror(in) || fread(buf, 1, size, in) != size ) {
        fprintf(stddiag(), "Error reading `%s'.\n", filename);
        free(buf);
        fclose(in);
        return NULL;
    }
    fclose(in);
    ((char*)buf)[size]= 0;
    if( filesize )
        *filesize= size;
    return buf;
}


/*!
    \brief Compare two strings, ignoring case unless they differ only by case.

    \param s1 First string to compare
    \param s2 Second string to compare

    \return <0 if s1 precedes s2 lexicographically, 0 if they are equal, >0
    otherwise
*/
int strcase2cmp(const void *s1, const void *s2)
{
    int result;

    result= strcasecmp(s1, s2);
    if( result )
        return result;
    return strcmp(s1, s2);
}


/*!
    \brief Compare a word, ignoring case

    Compares two strings up to the first non-alphanumeric character, case
    insensitively.

    \param s1, s2 strings to compare

    \return 0 if equal, <0 if differing char of s1 is smaller, >0 otherwise
*/
int wordcmp(const char *s1, const char *s2)
{
    while( isalnum(*s1) ) {
        if( tolower(*s1) != tolower(*s2) )
            return *s1 - *s2;
        ++s1;
        ++s2;
    }
    return isalnum(*s2) ? *s1 - *s2 : 0;
}


/*!
    \brief Search for string of words, ignoring case and non-word characters

    Search for a string of words (= contiguous strings of alphanumeric
    characters), ignoring case and not comparing characters separating the
    words.  The return value is a pointer to the first character after the last
    word.

    \param haystack String to search
    \param lcwords String to search for.  This may contain only alphanumeric
    characters and spaces, and no leading and trailing spaces.

    \return Pointer to first character after the last word of the string of
    words in the haystack string; or NULL if the phrase was not found.
*/
char *afterwords(const char *haystack, const char *lcwords)
{
    const char *pos, *cmp, *word;

    pos= haystack; 
    while( *pos )
    {
        while( *pos && !isalnum(*pos) )
            ++pos;
        for( cmp= pos, word= lcwords; 13; ++cmp, ++word ) {
            if( !*word && !isalnum(*cmp) )
                return (char*)cmp;
            if( *word == ' ' && !isalnum(*cmp) ) {
                while( *word == ' ' )
                    ++word;
                while( *cmp && !isalnum(*cmp) )
                    ++cmp;
            }
            if( tolower(*cmp) != *word )
                break;
        }
        while( isalnum(*pos) )
            ++pos;
    }
    return NULL;
}

