/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "serialreader.h"

sr_config config;

struct config_item {
    char *key;
    char type;
    void *data;
};

/*!
    \brief List of config file options
*/
struct config_item options[]= {
    "font0regular", '$', &config.fontfiles[0],
    "font0bold", '$', &config.fontfiles[1],
    "font0italic", '$', &config.fontfiles[2],
    "font0mono", '$', &config.fontfiles[3],
    "font1regular", '$', &config.fontfiles[4],
    "font1bold", '$', &config.fontfiles[5],
    "font1italic", '$', &config.fontfiles[6],
    "font1mono", '$', &config.fontfiles[7],
    "font2regular", '$', &config.fontfiles[8],
    "font2bold", '$', &config.fontfiles[8],
    "font2italic", '$', &config.fontfiles[10],
    "font2mono", '$', &config.fontfiles[11],
    "font3regular", '$', &config.fontfiles[12],
    "font3bold", '$', &config.fontfiles[13],
    "font3italic", '$', &config.fontfiles[14],
    "font3mono", '$', &config.fontfiles[15],
    "fontsize", '#', &config.fontsize,
    "defaultencoding", '$', &config.defaultencoding,
    "forceencoding", '$', &config.forceencoding,
    "startfile", '$', &config.start_file,
    "startposition", '#', &config.start_position,
    "margin", '#', &config.margin,
    "", ' ', NULL
};


/*!
    \brief Read line in config file and assign value according to key

    Search the option definition array for the key (first word) in the current
    line of the config file and assign the value to the variable referenced in
    the option definition.  The key must be followed by spaces or tabs and /
    or an optional equals sign before the value.  String values extend from the
    first non-space/tab after the key or equals sign to the last in the line.

    \param read Pointer to line in a configuration file, will be updated to
    point to the start of the next line or (after the last line) to the NUL
    byte at the end of the file.

    \return 1 if the key was found, 0 otherwise
*/
int assign_item(char **read)
{
    char *scan, *end;
    int ind, found;
    char pre0;

    scan= *read;
    while( *scan == ' ' || *scan == '\t' )
        ++scan;
    found= 0;
    for( ind= 0; *options[ind].key; ++ind )
        if( !wordcmp(scan, options[ind].key) ) {
            found= 1;
            break;
        }
    if( found ) {
        while( *scan && !isspace(*scan) && *scan != '=' )
            ++scan;
        while( *scan == ' ' || *scan == '\t' )
            ++scan;
        if( *scan == '=' ) {
            ++scan;
            while( isspace(*scan) )
                ++scan;
        }
        if( options[ind].type == '#' ) {
            *(int*)options[ind].data= strtol(scan, &scan, 0);
        }
        else if( options[ind].type == '$' ) {
            for( end= scan; *end && *end != '\n'; ++end );
            while( isspace(end[-1]) )
                --end;
            pre0= *end;
            *end= 0;
            *(char**)options[ind].data= scan;
            *read= pre0 ? end+1 : end;
            return 1;     /* return now because the '\n' may have been eaten */
        }
    }
    while( *scan && *scan != '\n' )
        ++scan;
    *read= *scan ? scan+1 : scan;
    return found;
}


/*!
    \brief Parse a configuration file

    assign_item() is called for every non-empty, non-comment line of the
    configuration file.  Comment lines contain optional white space and a '#'
    character followed by the comment.  Comments cannot occur after a key /
    value pair setting an option.

    \param conf Pointer to buffer containing the configuration file, concluded
    by a NUL character.
*/
void parse_config(char *conf)
{
    if( !conf )
        return;
    while( 13 )
    {
        while( isspace(*conf) )
            ++conf;
        if( !*conf )
            return;
        if( *conf == '#' ) {
            while( *conf && *conf != '\n' )
                ++conf;
            continue;
        }
        assign_item(&conf);
    }
}


/*!
    \brief Read global and local configuration file

    Read the config files into RAM and parse them.  The local configuration
    file is assumed to be located in the current directory.  The local settings
    override those from the global configuration file in case of conflicts.
    The fonts are rearranged so that the size range starts at zero and ends at
    the last font size where a regular font is available.  Every valid set of
    fonts has to include a regular font because it is the fallback for
    unavailable styles and glyphs.

    \return 0 if successful, -1 if no regular font file name was given anywhere
*/
int read_config(void)
{
    static char *globalconf, *localconf;

    int firstfont, endfont;

    config.fontsize= -1;
    config.margin= 5;
    free(localconf);
    free(globalconf);
    globalconf= load_file(GLOBAL_CONFIG, NULL);
    localconf= load_file(CONFIG_NAME, NULL);
    parse_config(globalconf);
    parse_config(localconf);    /* local settings override globals */
    for( firstfont= 0; firstfont < MAXFONTS && !config.fontfiles[firstfont];
                firstfont += NSTYLES );
    if( firstfont == MAXFONTS )
        return -1;
    for( endfont= firstfont + NSTYLES;
        endfont < MAXFONTS && config.fontfiles[endfont]; endfont += NSTYLES );
    memmove(config.fontfiles, config.fontfiles + firstfont,
            sizeof(char*) * (endfont - firstfont));
    config.numfontsizes= (endfont - firstfont) / NSTYLES;
    if( config.fontsize < 0 )
        config.fontsize= config.numfontsizes / 2;
    else if( config.fontsize >= config.numfontsizes )
        config.fontsize= config.numfontsizes - 1;
    return 0;
}


/*!
    \brief Write local configuration file with current file and position

    If a local configuration file exists, it is rewritten without the entries
    startfile and startposition, if present.  These entries with the updated
    position are then appended.

    \return 0 OK, !=0 file access error
*/
int write_config(void)
{
    FILE *conf;
    char *conffile, *scan, *chunkstart;
    unsigned chunklength;
    int status;

    conffile= load_file(CONFIG_NAME, NULL);
    conf= fopen(CONFIG_NAME, "w");
    if( !conf )
        return -1;
    if( conffile ) {
        /* First save existing settings except startfile, startposition and
           fontsize */
        chunkstart= conffile;
        scan= conffile;
        while( *scan ) {
            while( isspace(*scan) )
                ++scan;
            if( !wordcmp(scan, "startfile") || !wordcmp(scan, "startposition")
                    || !wordcmp(scan, "fontsize") ) {
                chunklength= scan - chunkstart;
                while( chunklength ) {
                    --chunklength;
                    if( chunkstart[chunklength] == '\n' ) {
                        ++chunklength;
                        break;
                    }
                }
                if( chunklength )
                    fwrite(chunkstart, 1, chunklength, conf);
                chunkstart= NULL;
            }
            while( *scan && *scan != '\n' )
                ++scan;
            if( *scan == '\n' )
                ++scan;
            if( !chunkstart )
                chunkstart= scan;
        }
        if( scan > chunkstart )
            fwrite(chunkstart, 1, scan - chunkstart, conf);
        free(conffile);
    }
    fprintf(conf, "startfile = %s\n", document->filename);
    fprintf(conf, "startposition = %u\n",
    (unsigned)(((text_pos*)document->lines.buf)[document->curr_first_line].pos
            - document->text) );
    fprintf(conf, "fontsize = %u\n", config.fontsize);
    status= ferror(conf);
    fclose(conf);
    return status;
}


