/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "serialreader.h"

/* Maximum length of encoding name: */
#define ENCODINGMAX 20
/* Maximum number of nested HTML tags affecting style (<b>, <i>, etc.): */
#define STYLEMAXDPETH 64

typedef struct {
    autospace text;
    char *file, *write;
    char encoding[ENCODINGMAX];
    unsigned *enctab;
    long filesize;
    int html, utf8;
    /* Elements used only for HTML: */
    char *read;
    char html_tag, html_pre, eat_space, have_space;
    char stylestack[STYLEMAXDPETH];
    int styledepth;
}
parse_data;

void append_string(autospace *dest, char **write, const char *source, unsigned length);
unsigned *find_encoding(const char *enc);
void detect_type(parse_data *pd);
void preprocess_plain(parse_data *pd);
void html_entity(parse_data *pd);
void html_ignore_upto(parse_data *pd, char c);
void html_tag_skip(parse_data *pd);
const char *findinstrary(const char *list, int length, const char *str);
void html_tag(parse_data *pd);
void preprocess_html(parse_data *pd);


/*!
    \brief Append string to extensible buffer

    Copies a string to a write position in a buffer while calling autoextend()
    as applicable.  The write pointer is updated, but the concluding zero byte
    i snot copied.  If a necessary autoexten() fails, the write pointer is left
    where it is, and the function returns.

    \param dest Pointer to destination autoextensible buffer
    \param write Write pointer into the destination buffer
    \param source Pointer to string to copy
    \param length Maximum length of string to copy, if NUL is not encountered
    first.  If length is 0, copy whole string.
*/
void append_string(autospace *dest, char **write, const char *source, unsigned length)
{
    unsigned ind, oldsize;

    ind= (*write - (char*)dest->buf);
    if( length ) {
        while( ind + length > dest->size ) {
            oldsize= dest->size;
            autoextend(dest);
            if( dest->size == oldsize )
                return;
        }
        length += ind;
        while( *source && ind < length )
            ((char*)dest->buf)[ind++]= *source++;
    }
    else {
        while( *source )
        {
            if( ind >= dest->size ) {
                autoextend(dest);
                if( ind >= dest->size )
                    return;
            }
            ((char*)dest->buf)[ind++]= *source++;
        }
    }
    *write= (char*)dest->buf + ind;
    /* may be different memory region than initially, due to autoextend() */
}


/*!
    \brief Find character encoding table by name

    Searches the global list of 8-bit encodings by name and returns the
    conversion table to UTF-8.  The table contains 128 unsigned values each of
    which is a UTF-8 encoded string up to 4 bytes long.

    \param enc Name of the requested 8-bit encoding

    \return Pointer to UTF-8 conversion table, or NULL if encoding was not found
*/
unsigned *find_encoding(const char *enc)
{
    encoding *scan;

    for( scan= encodings8bit; scan->name; ++scan )
        if( !strcasecmp(enc, scan->name) )
            return scan->utf8table;
    return NULL;
}


/*!
    \brief Work out file type and character encoding

    Tries to detect HTML by searching for <html> and <body> tags.  For HTML,
    the character encoding is determined from the content-type meta tag.  The
    UTF-8 "byte-order mark" is also recognised (for HTML and plain text).
    find_encoding() is called to obtain the encoding table.

    \param pd Pointer to parse data struct which contains pointer to file data
    and to which the file type and UTF-9 flag and the pointer to the conversion
    table will be written
*/
#define HTMLHEADMAX 2048

void detect_type(parse_data *pd)
{
    char *read;
    char save;

    if( pd->filesize > HTMLHEADMAX ) {
        save= pd->file[HTMLHEADMAX];
        pd->file[HTMLHEADMAX]= 0;
    }
    pd->html= strcasestr(pd->file, "<html>") || strcasestr(pd->file, "<html ")
         || strcasestr(pd->file, "<body>") || strcasestr(pd->file, "<body ");
    if( config.forceencoding && *config.forceencoding ) {
        if( !strcasecmp(config.forceencoding, "utf-8") )
            pd->utf8= 1;
        else {
            pd->utf8= 0;
            pd->enctab= find_encoding(config.forceencoding);
        }
        if( pd->filesize > HTMLHEADMAX )
            pd->file[HTMLHEADMAX]= save;
        return;
    }
    if( config.defaultencoding && *config.defaultencoding )
        if( !strcasecmp(config.defaultencoding, "utf-8") )
            pd->utf8= 1;
        else {
            pd->utf8= 0;
            pd->enctab= find_encoding(config.defaultencoding);
        }
    else {
        pd->utf8= 0;
        pd->enctab= encodings8bit[0].utf8table;
    }
    /* Set UTF-8 flag according to UTF-8 "byte order mark" */
    if(pd->file[0] == '\xEF' && pd->file[1] == '\xBB' && pd->file[2] == '\xBF')
        pd->utf8= 1;
    if( pd->html ) {
        read= afterwords(pd->file, "meta http equiv content type content text html charset");
        if( read && *read == '=' ) {
            strncpy(pd->encoding, read+1, ENCODINGMAX);
            pd->encoding[ENCODINGMAX]= 0;
            for( read= pd->encoding; read < pd->encoding+ENCODINGMAX; ++read )
                if( *read == '"' ) {
                    *read= 0;
                    break;
                }
            if( !strcasecmp(pd->encoding, "utf-8") )
                pd->utf8= 1;
            else {
                pd->utf8= 0;
                pd->enctab= find_encoding(pd->encoding);
            }
        }
    }
    if( pd->filesize > HTMLHEADMAX )
        pd->file[HTMLHEADMAX]= save;
}


/*------------------ Function for plain text processing -----------------*/

/*!
    \brief Convert a plain text file to internal UTF-8 representation

    Scans and transforms a plain text file and writes the result to the
    auto-extensible destination buffer.  Tabs are replaced by spaces, double
    newlines (with optional white space in between) are marked as paragraphs,
    characters outside ASCII range are converted if the encoding is known and
    marked as unknown otherwise.  Other control characters are deleted.

    \param pd Parse data struct containing pointer to file contents and
    autoextensible destination buffer.
*/
void preprocess_plain(parse_data *pd)
{
    char *chunkstart, *read, *nextutf8;
    int ind, lastindent, lastindent2, hadspace, havespace;
    char c;

    read= chunkstart= pd->file;
    lastindent= 0;
    lastindent2= 1;
    hadspace= 0;
    while( *read ) {
        havespace= 0;
        if( *read > 0x20 && *read < 0x7F )
            ++read;
        else if( *read == ' ' || *read == '\t' ) {
            if( !hadspace )
                *read++ = ' ';/* cut sequence of tabs and spaces to one space */
            while( *read && *read <= 0x20 && *read != '\n' )
                *read++ = CHAR_ERASED;
            havespace= 1;
        }
        else if( *read == '\n' ) {
            for( ind= 1; read[ind] && read[ind] <= 0x20; ++ind ) {
                c= read[ind];
                read[ind]= CHAR_ERASED;
                if( c == '\n' )
                    read[1]= CHAR_PARAGRAPH;
            }
            if( ind - 1 > lastindent &&
                    (lastindent == lastindent2 || ind - 1 == lastindent2) )
                read[1]= CHAR_PARAGRAPH;
            lastindent2= lastindent;
            lastindent= ind - 1;
            if( read[1] && read[1] != CHAR_PARAGRAPH ) {
                *read= hadspace? CHAR_ERASED : ' ';    /* fill paragraphs */
                havespace= 1;
            }
            else if( hadspace )
                read[-1]= CHAR_ERASED;
                /* spaces before newlines throw paragraph breaking off */
            read += ind;
        }
        else if( *(unsigned char*)read >= 0x80 ) {
            if( pd->utf8 )
                ++read;
            else if( pd->enctab ) {
                if( read > chunkstart )
                    append_string(&pd->text, &pd->write, chunkstart, read - chunkstart);
                append_string(&pd->text, &pd->write,
                        (char*)(pd->enctab + *(unsigned char*)read - 0x80), 4);
                chunkstart= ++read;
            }
            else if( valid_utf8(read, (const char **)&nextutf8) )
                read= nextutf8;
            else
                *read++ = CHAR_UNKNOWN;
        }
        else if( *(unsigned char*)read >= 0x80 && !pd->utf8 && pd->enctab ) {
            if( read > chunkstart )
                append_string(&pd->text, &pd->write, chunkstart, read - chunkstart);
            append_string(&pd->text, &pd->write,
                        (char*)(pd->enctab + *(unsigned char*)read - 0x80), 4);
            chunkstart= ++read;
        }
        else
            *read++ = CHAR_ERASED;
        hadspace=  havespace;
    }
    append_string(&pd->text, &pd->write, chunkstart, read - chunkstart + 1);
    *pd->write= 0;
}


/*------------------ Functions and data for HTML processing -----------------*/

struct named_entity {
    char name[10];
    int codepoint;
};

#define NAMED_ENTITIES  252
struct named_entity named_entities[NAMED_ENTITIES]= {
"AElig", 198, "Aacute", 193, "Acirc", 194, "Agrave", 192, "Alpha", 913,
"Aring", 197, "Atilde", 195, "Auml", 196, "Beta", 914, "Ccedil", 199, "Chi",
935, "Dagger", 8225, "Delta", 916, "ETH", 208, "Eacute", 201, "Ecirc", 202,
"Egrave", 200, "Epsilon", 917, "Eta", 919, "Euml", 203, "Gamma", 915, "Iacute",
205, "Icirc", 206, "Igrave", 204, "Iota", 921, "Iuml", 207, "Kappa", 922,
"Lambda", 923, "Mu", 924, "Ntilde", 209, "Nu", 925, "OElig", 338, "Oacute",
211, "Ocirc", 212, "Ograve", 210, "Omega", 937, "Omicron", 927, "Oslash", 216,
"Otilde", 213, "Ouml", 214, "Phi", 934, "Pi", 928, "Prime", 8243, "Psi", 936,
"Rho", 929, "Scaron", 352, "Sigma", 931, "THORN", 222, "Tau", 932, "Theta",
920, "Uacute", 218, "Ucirc", 219, "Ugrave", 217, "Upsilon", 933, "Uuml", 220,
"Xi", 926, "Yacute", 221, "Yuml", 376, "Zeta", 918, "aacute", 225, "acirc",
226, "acute", 180, "aelig", 230, "agrave", 224, "alefsym", 8501, "alpha", 945,
"amp", 38, "and", 8743, "ang", 8736, "aring", 229, "asymp", 8776, "atilde",
227, "auml", 228, "bdquo", 8222, "beta", 946, "brvbar", 166, "bull", 8226,
"cap", 8745, "ccedil", 231, "cedil", 184, "cent", 162, "chi", 967, "circ", 710,
"clubs", 9827, "cong", 8773, "copy", 169, "crarr", 8629, "cup", 8746, "curren",
164, "dArr", 8659, "dagger", 8224, "darr", 8595, "deg", 176, "delta", 948,
"diams", 9830, "divide", 247, "eacute", 233, "ecirc", 234, "egrave", 232,
"empty", 8709, "emsp", 8195, "ensp", 8194, "epsilon", 949, "equiv", 8801,
"eta", 951, "eth", 240, "euml", 235, "euro", 8364, "exist", 8707, "fnof", 402,
"forall", 8704, "frac12", 189, "frac14", 188, "frac34", 190, "frasl", 8260,
"gamma", 947, "ge", 8805, "gt", 62, "hArr", 8660, "harr", 8596, "hearts", 9829,
"hellip", 8230, "iacute", 237, "icirc", 238, "iexcl", 161, "igrave", 236,
"image", 8465, "infin", 8734, "int", 8747, "iota", 953, "iquest", 191, "isin",
8712, "iuml", 239, "kappa", 954, "lArr", 8656, "lambda", 955, "lang", 9001,
"laquo", 171, "larr", 8592, "lceil", 8968, "ldquo", 8220, "le", 8804, "lfloor",
8970, "lowast", 8727, "loz", 9674, "lrm", 8206, "lsaquo", 8249, "lsquo", 8216,
"lt", 60, "macr", 175, "mdash", 8212, "micro", 181, "middot", 183, "minus",
8722, "mu", 956, "nabla", 8711, "nbsp", 160, "ndash", 8211, "ne", 8800, "ni",
8715, "not", 172, "notin", 8713, "nsub", 8836, "ntilde", 241, "nu", 957,
"oacute", 243, "ocirc", 244, "oelig", 339, "ograve", 242, "oline", 8254,
"omega", 969, "omicron", 959, "oplus", 8853, "or", 8744, "ordf", 170, "ordm",
186, "oslash", 248, "otilde", 245, "otimes", 8855, "ouml", 246, "para", 182,
"part", 8706, "permil", 8240, "perp", 8869, "phi", 966, "pi", 960, "piv", 982,
"plusmn", 177, "pound", 163, "prime", 8242, "prod", 8719, "prop", 8733, "psi",
968, "quot", 34, "rArr", 8658, "radic", 8730, "rang", 9002, "raquo", 187,
"rarr", 8594, "rceil", 8969, "rdquo", 8221, "real", 8476, "reg", 174, "rfloor",
8971, "rho", 961, "rlm", 8207, "rsaquo", 8250, "rsquo", 8217, "sbquo", 8218,
"scaron", 353, "sdot", 8901, "sect", 167, "shy", 173, "sigma", 963, "sigmaf",
962, "sim", 8764, "spades", 9824, "sub", 8834, "sube", 8838, "sum", 8721,
"sup", 8835, "sup1", 185, "sup2", 178, "sup3", 179, "supe", 8839, "szlig", 223,
"tau", 964, "there4", 8756, "theta", 952, "thetasym", 977, "thinsp", 8201,
"thorn", 254, "tilde", 732, "times", 215, "trade", 8482, "uArr", 8657,
"uacute", 250, "uarr", 8593, "ucirc", 251, "ugrave", 249, "uml", 168, "upsih",
978, "upsilon", 965, "uuml", 252, "weierp", 8472, "xi", 958, "yacute", 253,
"yen", 165, "yuml", 255, "zeta", 950, "zwj", 8205, "zwnj", 8204 };

/*!
    \brief Translate an HTML entity into unicode

    Parses an HTML entity and writes the the corresponding unicode character to
    the text.  The read pointer is advanced to after the entity.  The
    concluding semicolon is skipped if present, but its absence is ignored.  An
    unknown named entity is transcribed to the unknown character marker.

    \param pd Parse data struct, with the read pointer positioned after the '&'
    indicating the entity
*/
void html_entity(parse_data *pd)
{
    struct named_entity *this_ent;
    char *scan;
    int codepoint, ind;
    char save;

    scan= pd->read;
    if( *scan == '#' ) {
        if( (scan)[1] == 'x' || scan[1] == 'X' )
            codepoint= strtol(scan+2, &scan, 16);
        else
            codepoint= strtol(scan+1, &scan, 10);
        if( *scan == ';' )
            ++scan;
    }
    else {
        while( isalpha(*scan) ) {
            *scan= tolower(*scan);
            ++scan;
        }
        save= *scan;
        *scan= 0;
        this_ent= bsearch(pd->read, named_entities, NAMED_ENTITIES,
                          sizeof(struct named_entity),
                          (int (*)(const void *, const void *))strcmp);
        codepoint= this_ent ? this_ent->codepoint : CHAR_UNKNOWN;
        *scan= save;
        if( *scan == ';' )
            ++scan;
    }
    pd->read= scan;
    if( pd->write > (char*)pd->text.buf + pd->text.size - 4 ) {
        ind= pd->write - (char*)pd->text.buf;
        autoextend(&pd->text);
        pd->write= (char*)pd->text.buf + ind;
    }
    utf8_write(codepoint, pd->write, &pd->write);
}


/*!
    \brief Skip HTML until after a given character

    Comments and quoted strings in tags are skipped even if they contain the
    requested character.  If the character is not found, the read position is
    set to the concluding NUL byte.

    \param read Pointer to pointer to read position in HTML text
    \param c Character to search for
*/
void html_ignore_upto(parse_data *pd, char c)
{
    char *read= pd->read, *end;

    while( *read && *read != c ) {
        if( pd->html_tag && *read == '"' ) {
            ++read;
            while( *read && *read != '"' )
                ++read;
            ++read;
        }
        else if( !pd->html_tag && *read == '<' && read[1] == '!' &&
                 read[2] == '-' && read[3] == '-' ) {
            end= strstr(read + 4, "-->");
            if( end )
                read= end + 3;
            else
                read += 4;
        }
        else
            ++read;
    }
    if( *read )
        ++read;
    pd->read= read;
}


/*!
    \brief Skip an HTML tag including its contents

    Skips HTML from a tag up to a corresponding closing tag.  Nested tags are
    not tracked; there is assumed to be no nesting of the given tag.

    \param pd Parse data struct, with the read pointer positioned after the '<'
    initiating the tag, and html_tag true
*/
void html_tag_skip(parse_data *pd)
{
    char *tag;
    int depth;

    tag= pd->read;
    depth= 1;
    html_ignore_upto(pd, '>');
    pd->html_tag= 0;
    while( *pd->read && depth )
    {
        html_ignore_upto(pd, '<');
        if( *pd->read == '/' )
            if( !wordcmp(pd->read+1, tag) )
                --depth;
            else;
        else if( !isalpha(*pd->read) && *pd->read != '!' ) {
            ++pd->read;
            continue;       /* assume this is no tag */
        }
        else if( !wordcmp(pd->read, tag) )
            ++depth;
        pd->html_tag= 1;
        html_ignore_upto(pd, '>');
        pd->html_tag= 0;
    }
}


#define TAGMAXLEN   10

struct style_tag {
    char tag[8];
    char style;
};

const struct style_tag styletags[]= { "i", STYLE_ITALIC, "b", STYLE_BOLD, "tt",
STYLE_MONO, "em", STYLE_ITALIC, "strong", STYLE_BOLD, "code", STYLE_MONO, "" };
const char zerowidthtags[][TAGMAXLEN]= {
"u", "strike", "s", "big", "small", "samp", "kbd", "var", "cite", "dfn", "div",
"abbr", "acronym", "q", "a", "blink", "font", "basefont", "wbr", "nobr", "bdo",
"" };
const char skiptags[][TAGMAXLEN]= { "head", "script", "form", "style", "" };
const char xformtags[][TAGMAXLEN]= {
"p", { '\n', CHAR_PARAGRAPH, 0 }, "sub", " ^ ", "sup", " _ ", "br", "\n",
"h1", { '\n', CHAR_STOP, '\n', STYLE_BOLD, 0 },
"h2", { '\n', CHAR_STOP, '\n', STYLE_BOLD, 0 },
"h3", { '\n', CHAR_STOP, '\n', STYLE_BOLD, 0 },
"h4", { '\n', CHAR_STOP, '\n', 0 }, "h5", { '\n', CHAR_ERASED, '\n', 0 },
"dt", { '\n', STYLE_BOLD, 0 }, "li", "\n\xE2\x80\xA2 ", "" };
const char xformetags[][TAGMAXLEN]= {
"p", { '\n', CHAR_PARAGRAPH, 0 }, "tr", "\n", "td", " | ",
"h1", { '\n', STYLE_REGULAR, 0 }, "h2", { '\n', STYLE_REGULAR, 0 },
"h3", { '\n', STYLE_REGULAR, 0 }, "h4", "\n", "h5", "\n",
"dt", { '\n', STYLE_REGULAR, 0 }, "dl", "\n", "ul", "\n", "ol", "\n",
"dd", "\n", "" };

/*!
    \brief Find a string in an array of strings

    Searches for a string in an array of strings of equal length.

    \param list Array of strings, concluded by empty string
    \param length Length of each string in the array
    \param str String to search for

    \return Pointer to string array element matching search string
*/
const char *findinstrary(const char *list, int length, const char *str)
{
    while( *list ) {
        if( !strcmp(list, str) )
            return list;
        list += length;
    }
    return NULL;
}


/*!
    \brief Process an HTML tag

    Decides what to do with an HTML tag based on the lists of zero-width
    tags, tags to be transcribed, and tags to be skipped including content.
    The default action is to replace a tag by a single space.

    \param pd Parse data struct, with the read pointer positioned after the '<'
    initiating the tag
*/
void html_tag(parse_data *pd)
{
    char *scan, *end;
    const char *foundtag;
    int endtag;
    char save, style;

    if( *pd->read == '!' ) {
        if( pd->read[1] == '-' && pd->read[2] == '-' ) {
            end= strstr(pd->read + 3, "-->");
            if( end )
                pd->read= end + 3;
            else
                pd->read += 3;
        }
        else    /* document type definition or some such */
            html_ignore_upto(pd, '>');
        pd->have_space= pd->eat_space;
    }
    else {
        endtag= *pd->read == '/';
        if( endtag )
            ++pd->read;
        scan= pd->read;
        while( isalnum(*scan) ) {
            *scan=  tolower(*scan);
            ++scan;
        }
        save= *scan;
        *scan= 0;
        if( !strcmp(pd->read, "pre") )
            pd->html_pre= !endtag;
        if( foundtag= findinstrary((const char*)styletags, sizeof(struct style_tag), pd->read) ) {
            style= 0;
            if( endtag ) {
                if( pd->styledepth ) {
                    --pd->styledepth;
                    if( !pd->styledepth )
                        style= STYLE_REGULAR;
                    else if( pd->styledepth <= STYLEMAXDPETH )
                        style= pd->stylestack[pd->styledepth - 1];
                }
            }
            else {
                if( pd->styledepth < STYLEMAXDPETH ) {
                    style= ((struct style_tag*)foundtag)->style;
                    pd->stylestack[pd->styledepth]= style;
                }
                ++pd->styledepth;
            }
            if( style )
                append_string(&pd->text, &pd->write, &style, 1);
            pd->have_space= pd->eat_space;
        }
        else if( foundtag= endtag ?
                        findinstrary(xformetags[0], 2*TAGMAXLEN, pd->read) :
                        findinstrary(xformtags[0], 2*TAGMAXLEN, pd->read) ) {
            append_string(&pd->text, &pd->write, foundtag+TAGMAXLEN, 0);
            pd->have_space= pd->write[-1] <= ' ';
        }
        else if( !endtag && findinstrary(skiptags[0], TAGMAXLEN, pd->read) ) {
            *scan= save;
            html_tag_skip(pd);
            pd->have_space= pd->eat_space;
            return;
        }
        else if( !findinstrary(zerowidthtags[0], TAGMAXLEN, pd->read) ) {
            if( !pd->eat_space )
                append_string(&pd->text, &pd->write, " ", 1);
            pd->have_space= 1;
        }
        *scan= save;
        pd->read= scan;
        html_ignore_upto(pd, '>');
    }
}


/*!
    \brief Preprocess an HTML file

    The text is converted according to the character encoding (if known),
    paragraphs are marked and tags changing the style are converted.  Other
    tags are removed or transformed as appropriate.  This function calls
    html_entity() to handle entities and html_tag() to handle tags.

    \param pd Pointer to parse data struct
*/
void preprocess_html(parse_data *pd)
{
    char *chunkstart, *nextutf8;

    pd->html_tag= 0;
    pd->html_pre= 0;
    pd->eat_space= 0;
    pd->read= pd->file;
    pd->styledepth= 0;

    chunkstart= pd->read;
    while( *pd->read )
    {
        pd->have_space= 0;
        if( *pd->read == '<' && (isalpha(pd->read[1]) || pd->read[1] == '/'
                                 || pd->read[1] == '!') ) {
            if( pd->read > chunkstart )
                append_string(&pd->text, &pd->write, chunkstart, pd->read - chunkstart);
            ++pd->read;
            pd->html_tag= 1;
            html_tag(pd);
            pd->html_tag= 0;
            chunkstart= pd->read;
        }
        else if( *pd->read == '&' && (isalpha(pd->read[1]) || pd->read[1] == '#' ) ) {
            if( pd->read > chunkstart )
                append_string(&pd->text, &pd->write, chunkstart, pd->read - chunkstart);
            ++pd->read;
            html_entity(pd);
            chunkstart= pd->read;
        }
        else if( *(unsigned char *)pd->read >= 0x80 ) {
            if( pd->utf8 )
                ++pd->read;
            else if( pd->enctab ) {
                if( pd->read > chunkstart )
                    append_string(&pd->text, &pd->write, chunkstart, pd->read - chunkstart);
                append_string(&pd->text, &pd->write, (char*)(pd->enctab +
                                *(unsigned char*)pd->read - 0x80), 4);
                chunkstart= ++pd->read;
            }
            else if( valid_utf8(pd->read, (const char **)&nextutf8) )
                pd->read= nextutf8;
            else
                *pd->read++ = CHAR_UNKNOWN;
        }
        else if( isspace(*pd->read) && (!pd->html_pre || *pd->read != '\n') ) {
            *pd->read++ = pd->eat_space ? CHAR_ERASED : ' ';
            pd->have_space= 1;
        }
        else
            ++pd->read;
        pd->eat_space= pd->have_space;
    }
    append_string(&pd->text, &pd->write, chunkstart, pd->read - chunkstart + 1);
    *pd->write= 0;
}


/*!
    \brief Load a new document

    Loads a document (plain text or HTML) file into RAM, determines its type
    and encoding, converts it to the internal representation, creates an
    sr_document struct and performs line breaking.

    \param filename Document file name

    \return Pointer to sr_document struct containing pointers to other data.
    Must be freed by discard_document().
*/
sr_document *load_document(char *filename)
{
    parse_data pd;
    sr_document *docheader;
    char *read, *write;

    pd.file= load_file(filename, &pd.filesize);
    if( !pd.file )
        return NULL;
    pd.text.size= pd.filesize + sizeof(sr_document);
    pd.text.increment= 1024;
    autoalloc(&pd.text);
    if( !pd.text.buf ) {
        free(pd.file);
        return NULL;
    }
    pd.write= (char*)pd.text.buf + sizeof(sr_document);
    detect_type(&pd);
    if( pd.html )
        preprocess_html(&pd);
    else
        preprocess_plain(&pd);
    free(pd.file);
    docheader= pd.text.buf;
    docheader->text= (char*)docheader + sizeof(sr_document);
    docheader->lines.buf= NULL;
    docheader->lines.elemsize= sizeof(text_pos);
    docheader->lines.maxelem= pd.filesize / 100;
    break_document(docheader->text, &docheader->lines);
    docheader->curr_first_line= 0;
    return docheader;
}


/*!
    \brief Deallocate a document and auxiliary data

    This function must be used to free the struct returned by load_document().

    \param doc Pointer to document struct (NULL is ignored)
*/
void discard_document(sr_document *doc)
{
    if( !doc )
        return;
    free(doc->lines.buf);
    free(doc);
}


