/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

/*
    Functions for reading bitmap glyphs embedded in otf fonts.

    Conveniently enough for testing, the machine parameters of the ARM in the
    Iliad are the same as those of desktop processors.
*/

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include "serialreader.h"

#pragma pack(1)

/*  Packed OTF structs.  ALL DATA TYPES ARE BIG-ENDIAN!  */

typedef struct {
    unsigned version;           /* 0x00010000 big-endian */
    unsigned fontRevision;
    unsigned checkSumAdjustment;
    unsigned magicNumber;       /* 0x5F0F3CF5 big-endian */
    ushort flags;
    ushort unitsPerEm;
    long long created;
    long long modified;
    short xMin;
    short yMin;
    short xMax;
    short yMax;
    ushort macStyle;
    ushort lowestRecPPEM;
    short fontDirectionHint;
    short indexToLocFormat;
    short glyphDataFormat;
}
otf_header;

typedef struct {
    schar ascender;
    schar descender;
    uchar widthMax;
    schar caretSlopeNumerator;
    schar caretSlopeDenominator;
    schar caretOffset;
    schar minOriginSB;
    schar minAdvanceSB;
    schar maxBeforeBL;
    schar minAfterBL;
    schar pad1;
    schar pad2;
}
sbitLineMetrics;

typedef struct {
    unsigned indexSubTableArrayOffset;
    unsigned indexTablesSize;
    unsigned numberOfIndexSubTables;
    unsigned colorRef;
    sbitLineMetrics hori;
    sbitLineMetrics vert;
    ushort startGlyphIndex;
    ushort endGlyphIndex;
    uchar ppemX;
    uchar ppemY;
    uchar bitDepth;
    schar flags;
}
bitmapSizeTable;

typedef struct {
    ushort firstGlyphIndex;
    ushort lastGlyphIndex;
    unsigned additionalOffsetToIndexSubtable;
}
indexSubTableArray;

typedef struct {
    ushort indexFormat;
    ushort imageFormat;
    unsigned imageDataOffset;
}
indexSubHeader;

/* smallGlyphMetrics defined in serialreader.h */

typedef struct {
    uchar height;
    uchar width;
    schar horiBearingX;
    schar horiBearingY;
    uchar horiAdvance;
    schar vertBearingX;
    schar vertBearingY;
    uchar vertAdvance;
}
bigGlyphMetrics;

#pragma pack()

typedef struct {
    void *table, *top;
}
otf_table;

/* otb_glyph defined in serialreader.h */

typedef struct {
    void *buf;
    otb_glyph *glyphs;
    otf_header *head;
    otf_table cmap, EBDT, EBLC;
    int cmapsubtype, cmap4size;
    unsigned *cmap12subtab, *cmap12top, *cmap12prevsegment;
    unsigned short *cmap4end, *cmap4prevsegment;
}
otf_font;


smallGlyphMetrics no_glyph_metrics = { 15, 10, 1, 12, 12 };
otb_glyph no_glyph_struct = {
    &no_glyph_metrics,
    "\x3F\xFF\xC0" "\xFF\xFF\xF0" "\xF0\x00\xF0" "\xF1\xA4\xF0"
    "\xF2\xA8\xF0" "\xF2\x08\xF0" "\xF0\x08\xF0" "\xF0\x68\xF0"
    "\xF0\xA0\xF0" "\xF0\x00\xF0" "\xF0\xA0\xF0" "\xF0\xA0\xF0"
    "\xF0\x00\xF0" "\xFF\xFF\xF0" "\x3F\xFF\xC0",
    45, 0, 12
};

const otb_glyph *no_glyph= &no_glyph_struct;


/*---------------------- OTF font reading functions ---------------------*/

#define OTF_ILLEGAL         (-1)
#define OTF_UNSUPPORTED     (-2)
#define OTF_UNSUCCESSFUL    (-3)


#define TAGEQ(a, b)     (*(unsigned*)(a) == *(unsigned*)(b))

/*!
    \brief Find relevant OTF tables in OTF font data

    Finds OTF tables needed by other functions and writes pointers to them to
    the font struct.  This does not need to be called explicitly, as the
    functions which need those pointers call this function when needed.

    \param font Font data structure

    \return 0 on success, or one of the error values OTF_UNSUCCESSFUL (not all
    tables were present) or OTF_ILLEGAL (illegal magic or version value)
*/
int otf_find_tables(otf_font *font)
{
    unsigned *tabrecord;
    int ntables, tabnr;

    if( !TAGEQ(font->buf, "\x00\x01\x00\x00") )
        return OTF_ILLEGAL;
    ntables= ntohs(((ushort *)font->buf)[2]);
    tabrecord= (unsigned *)font->buf + 3;
    for( tabnr= 0; tabnr< ntables; ++tabnr, tabrecord += 4 )
        if( TAGEQ(tabrecord, "head") )
            font->head= (otf_header*)((char*)font->buf + ntohl(tabrecord[2]));
        else if( TAGEQ(tabrecord, "cmap") ) {
            font->cmap.table= (char*)font->buf + ntohl(tabrecord[2]);
            font->cmap.top= (char*)font->buf + ntohl(tabrecord[2]) + ntohl(tabrecord[3]);
        }
        else if( TAGEQ(tabrecord, "EBDT") ) {
            font->EBDT.table= (char*)font->buf + ntohl(tabrecord[2]);
            font->EBDT.top= (char*)font->buf + ntohl(tabrecord[2]) + ntohl(tabrecord[3]);
        }
        else if( TAGEQ(tabrecord, "EBLC") ) {
            font->EBLC.table= (char*)font->buf + ntohl(tabrecord[2]);
            font->EBLC.top= (char*)font->buf + ntohl(tabrecord[2]) + ntohl(tabrecord[3]);
        }
    return font->cmap.table && font->EBDT.table && font->EBLC.table ?
            0 : OTF_UNSUCCESSFUL;
}


/*!
    \brief Find glyph index for a unicode code point

    Searches subtables of the cmap table for a given unicode code point.
    Currently only subtables with platform ID 0 and encoding ID 4 or platform
    ID 3 and encoding ID 10, in conjunction with an index table in format 12,
    are supported.

    \param font Font data structure
    \param codepoint Unicode code point

    \return 0 on success, or one of the error values OTF_UNSUCCESSFUL (code
    point not found), OTF_ILLEGAL (illegal magic or version value) or
    OTF_UNSUPPORTED (no format 12 subtable of cmap table found)
*/
int otf_glyph_index(otf_font *font, int codepoint)
{
    unsigned *cmap, *cmap12;
    unsigned short *cmap4;
    char *cmapsubtable;
    int status, max, ind, top, glyphind;

    if( !(font->cmapsubtype == 12 && font->cmap12subtab ||
          font->cmapsubtype == 4 && font->cmap4end) ) {
        if( !font->cmap.table ) {
            status= otf_find_tables(font);
            if( status < 0 )
                return status;
        }
        cmap= font->cmap.table;
        if( *(short*)cmap )
            return OTF_ILLEGAL;
        max= 2 * ntohs(((ushort*)cmap)[1]);
        for( ind= 1; ind < max && !TAGEQ(cmap+ind, "\x00\x00\x00\x04") &&
                     !TAGEQ(cmap+ind, "\x00\x03\x00\x0A"); ind += 2 );
        if( ind >= max )
            for( ind= 1; ind < max && !TAGEQ(cmap+ind, "\x00\x00\x00\x03") &&
                     !TAGEQ(cmap+ind, "\x00\x03\x00\x01"); ind += 2 );
        if( ind >= max )
            return OTF_UNSUPPORTED;
        cmapsubtable= (char*)font->cmap.table + ntohl(cmap[ind+1]);
        if( *cmapsubtable != 0 || (cmapsubtable[1] != 12 && cmapsubtable[1] != 4) )
            return OTF_UNSUPPORTED;
        font->cmapsubtype= cmapsubtable[1];
        if( font->cmapsubtype == 12 ) {
            font->cmap12subtab= (unsigned*)cmapsubtable + 4;
            font->cmap12top= font->cmap12subtab + 3 * ntohl(((unsigned*)cmapsubtable)[3]);
            font->cmap12prevsegment= NULL;
            for( cmap12= font->cmap12subtab; cmap12< font->cmap12top; ++cmap12 )
                *cmap12 = ntohl(*cmap12);
        }
        else {
            cmap4= (unsigned short *)cmapsubtable;
            top= ntohs(cmap4[1]) / 2;
            for( ind= 0; ind < top; ++ind )
                cmap4[ind]= ntohs(cmap4[ind]);
            font->cmap4size= cmap4[3] / 2;
            font->cmap4end= cmap4 + 7;
            font->cmap4prevsegment= NULL;
        }
    }
    if( font->cmapsubtype == 12 ) {
        if( font->cmap12prevsegment && font->cmap12prevsegment[0] <= codepoint
                && font->cmap12prevsegment[1] >= codepoint )
            return font->cmap12prevsegment[2] + codepoint - font->cmap12prevsegment[0];
        cmap12= font->cmap12subtab;
        while( cmap12 < font->cmap12top && cmap12[1] < codepoint )
            cmap12 += 3;
        if( cmap12 >= font->cmap12top || codepoint < cmap12[0] )
            return OTF_UNSUCCESSFUL;
        font->cmap12prevsegment= cmap12;
        return cmap12[2] + codepoint - cmap12[0];
    }
    else {
        if( font->cmap4prevsegment && font->cmap4prevsegment[0] >= codepoint
                && font->cmap4prevsegment[font->cmap4size+1] <= codepoint )
            cmap4= font->cmap4prevsegment;
        else {
            for( ind= 0; ind < font->cmap4size-1 &&
                         codepoint > font->cmap4end[ind]; ++ind );
            if( ind >= font->cmap4size-1 )
                return OTF_UNSUCCESSFUL;
            cmap4= font->cmap4end + ind;
            if( codepoint < cmap4[font->cmap4size+1] )
                return OTF_UNSUCCESSFUL;
            font->cmap4prevsegment= cmap4;
        }
        if( cmap4[3*font->cmap4size+1] ) {
            glyphind= cmap4[3*font->cmap4size+1
                            + codepoint - cmap4[font->cmap4size+1]
                            + cmap4[3*font->cmap4size+1] / 2];
            if( !glyphind )
                return OTF_UNSUCCESSFUL;
            else
                return glyphind + cmap4[2*font->cmap4size+1];
        }
        else {
            glyphind= (codepoint + cmap4[2*font->cmap4size+1]) % 0x10000;
            return glyphind ? glyphind : OTF_UNSUCCESSFUL;
        }
    }
}


/*!
    \brief Generate table of otb_glyph structs

    Generates a table containing structs with relevant information on every
    glyph in the font, such as advance value (width) and pointers to metrics
    structs and the bitmap.  Only the first strike of the font is considered.
    Only formats 1 to 3 of the index subtables in the EBLC table and formats 1,
    2, 5, 6 and 7 of the bitmaps in the EBDT table are supported.

    \param font Font data structure

    \return 0 on success, or one of the error values OTF_UNSUCCESSFUL (code
    point not found), OTF_ILLEGAL (illegal magic or version value) or
    OTF_UNSUPPORTED (unsupported subtable format encountered)
*/
int otb_list_glyphs(otf_font *font)
{
    char *loctab, *bittab;
    bitmapSizeTable *strike;
    indexSubTableArray *range;
    indexSubHeader *formats;
    int status, nranges, rangeind, indformat, bitformat, bitoffset, nextbitoffset;
    int glyph, firstglyph, lastglyph, endglyph, glyphoff;

    if( !font->EBLC.table && !font->EBDT.table ) {
        status= otf_find_tables(font);
        if( status < 0 )
            return status;
    }
    loctab= font->EBLC.table;
    if( !TAGEQ(loctab, "\x00\x02\x00\x00") )
        return OTF_ILLEGAL;
    strike= (bitmapSizeTable*)(loctab+8);
    if( strike->bitDepth != 2 || strike->flags != 1 )
        return OTF_UNSUPPORTED;
    endglyph= htons(strike->endGlyphIndex);
    font->glyphs= malloc( sizeof(otb_glyph) * (endglyph + 1) );
    if( ! font->glyphs )
        return OTF_UNSUCCESSFUL;
    memset( font->glyphs, 0, sizeof(otb_glyph) * (endglyph + 1) );
    range= (indexSubTableArray*)(loctab + ntohl(strike->indexSubTableArrayOffset));
    nranges= ntohl(strike->numberOfIndexSubTables);
    for( rangeind= 0; rangeind< nranges; ++rangeind, ++range )
    {
        formats= (indexSubHeader*)(loctab +
                        ntohl(strike->indexSubTableArrayOffset) +
                        ntohl(range->additionalOffsetToIndexSubtable));
        bitformat= ntohs(formats->imageFormat);
        indformat= ntohs(formats->indexFormat);
        if( (indformat < 1 || indformat > 3) || (bitformat != 1 && 
                    bitformat != 2 && bitformat < 5 && bitformat > 7) )
            return OTF_UNSUPPORTED;
        bittab= font->EBDT.table + ntohl(formats->imageDataOffset);
        firstglyph= ntohs(range->firstGlyphIndex);
        lastglyph= ntohs(range->lastGlyphIndex);
//  printf("range %d:  %d glyphs, index format %d\n", rangeind, lastglyph-firstglyph+1, indformat);   // DEBUG
        if( lastglyph > endglyph )
            lastglyph= endglyph;
        if( indformat == 2 ) {  /* same metrics & size for all glyphs in range */
            for( glyph= firstglyph, glyphoff= 0; glyph <= lastglyph; ++glyph, ++glyphoff ) {
                font->glyphs[glyph].bitpacked= bitformat != 1 && bitformat != 6;
                font->glyphs[glyph].bitsize= ntohl(((unsigned*)formats)[2]);
                font->glyphs[glyph].metrics= (smallGlyphMetrics*)((char*)formats + 12);
                font->glyphs[glyph].advance= font->glyphs[glyph].metrics->Advance;
                font->glyphs[glyph].bitmap= bittab + glyphoff * font->glyphs[glyph].bitsize;
            }
        }
        else {
            nextbitoffset= indformat == 1 ? ntohl(((unsigned*)formats)[2]) :
                                            ntohs(((ushort*)formats)[4]);
            for( glyph= firstglyph, glyphoff= 0; glyph <= lastglyph; ++glyph, ++glyphoff ) {
                font->glyphs[glyph].bitpacked= bitformat != 1 && bitformat != 6;
                bitoffset= nextbitoffset;
                font->glyphs[glyph].metrics= (smallGlyphMetrics*)(bittab + bitoffset);
                font->glyphs[glyph].advance= font->glyphs[glyph].metrics->Advance;
                bitoffset += bitformat < 5 ? SIZEOF_smallGlyphMetrics :
                                                  sizeof(bigGlyphMetrics);
                font->glyphs[glyph].bitmap= bittab + bitoffset;
                nextbitoffset= indformat == 1 ?
                                    ntohl(((unsigned*)formats)[3 + glyphoff]) :
                                    ntohs(((ushort*)formats)[5 + glyphoff]);
                font->glyphs[glyph].bitsize=  nextbitoffset - bitoffset;
            }
        }
    }
    return 0;
}


/*---------------------- High-level functions ------------------------*/


/*!
    \brief Load font file and initialise all font structs

    Loads the font file using load_file() and calls otb_list_glyphs() to
    generate the glyph table (which in turn calls otf_find_tables() to
    initialise pointers to important OTF tables).  The result must not be
    free()ed by the caller, because some pointed-to data has to be freed first.

    \param filename OTB font file name (OTF font with embedded bitmap font)
    
    \return Pointer to fully initialised font struct.  Not to be freed by the
    caller.
*/
otf_font *load_font(const char *filename)
{
    otf_font *font;
    int status;

    if( !filename || !*filename )
        return NULL;
    font= malloc(sizeof(otf_font));
    if( font ) {
        memset(font, 0, sizeof(otf_font));
        font->buf= load_file(filename, NULL);
    }
    if( !font || !font->buf ) {
        free(font);
        return NULL;
    }
    if( (status= otb_list_glyphs(font)) ) {
        free(font->buf);
        free(font);
        return NULL;
    }
    return font;
}


/*!
    \brief Deallocate font data.

    Frees memory used by font struct and data it points to.

    \param font Font struct to discard.
*/
void unload_font(otf_font *font)
{
    free(font->glyphs);
    free(font->buf);
    free(font);
}


static otf_font *fontset[NSTYLES];
static int typglypheight, typglyphdepth, parskip_;

/*!
    \brief Load new set of fonts

    This function loads the current set of fonts according to the font file
    name list and the current font size number in the global configuration
    struct.  It first tries to load the new regular font, which is a fallback
    for missing glyphs as well as missing bold, italic and mono fonts.  If the
    new regular font could not be loaded, the function gives up, otherwise it
    tries to load the fonts for the other styles.  Then the typical glyph
    height and depth and paragraph skip are determined based on the new regular
    font.

    \return 0 on success; 1 if the new regular font could not be loaded but a
    previous font set was available and was retained; -1 if no fonts are
    available
*/
int load_all_fonts(void)
{
    text_pos typstr= { "KTQfgy", STYLE_REGULAR };
    otf_font *newregularfont;
    int fontind, allok;

    newregularfont= load_font(config.fontfiles[config.fontsize*NSTYLES]);
    if( !newregularfont ) {
        fprintf(stddiag(), "Could not load %sregular font\n", fontset[0] ? "new ":"");
        return fontset[0] ? 1 : -1;
    }
    for( fontind= 0; fontind< NSTYLES; ++fontind )
        if( fontset[fontind] )
            unload_font(fontset[fontind]);
    fontset[0]= newregularfont;
    allok= 1;
    for( fontind= 1; fontind< NSTYLES; ++fontind ) {
        fontset[fontind]=
            load_font(config.fontfiles[config.fontsize*NSTYLES+fontind]);
        if( !fontset[fontind] )
            if( allok ) {
                allok= 0;
                fprintf(stddiag(), "Could not load fonts for style(s) %d", fontind);
            }
            else
                fprintf(stddiag(), ", %d", fontind);
    }
    if( !allok )
        fprintf(stddiag(), "\n");
    typglypheight= typglyphdepth= 0;
    line_height(&typstr, &typglypheight, &typglyphdepth, NULL);
    parskip_= (typglypheight + typglyphdepth) / 2;
    return 0;
}


/*!
    \brief Obtain a pointer to a glyph for a unicode code point in a given style

    \param codepoint Unicode code point
    \param style Writing style, translated to one of the loaded fonts

    \return Pointer to the corresponding otb_glyph struct.  If no glyph was
    found in the loaded fonts, or if the code point was CHAR_UNKNOWN, the
    fallback "no glyph" glyph is returned.
*/
const otb_glyph *get_glyph(int codepoint, int style)
{
    otf_font *fn;
    int glind;

    if( codepoint == CHAR_UNKNOWN )
        return no_glyph;
    if( style >= STYLE_REGULAR + NSTYLES || !fontset[style - STYLE_REGULAR] ) {
        fn= fontset[0];
        style= STYLE_REGULAR;
    }
    else
        fn= fontset[style - STYLE_REGULAR];
    glind= otf_glyph_index(fn, codepoint);
    if( style != STYLE_REGULAR && (glind < 0 || !fn->glyphs[glind].metrics) ) {
        fn= fontset[0];                 /* fall back to regular font */
        glind= otf_glyph_index(fn, codepoint);
    }
    if( glind < 0 || !fn->glyphs[glind].metrics )
        return no_glyph;
    else
        return fn->glyphs + glind;
}


/*!
    \brief Find out whether we have a glyph for a code point

    Return whether a glyph for a code point is present in the regular font of
    the current fontset.  Because the other fonts are not expected to contain
    more glyphs than the regular one and the regular font serves as a fallback,
    the others are not considered.

    \param codepoint Unicode code point

    \return 1 yes, 0 no
*/
int have_glyph(int codepoint)
{
    int glind;

    glind= otf_glyph_index(fontset[0], codepoint);
    return glind >= 0 && fontset[0]->glyphs[glind].metrics;
}


/*!
    \brief Determine line height and depth relative to baseline

    Determines the maximum height (over the baseline) and depth (under the
    baseline) of any glyph in the line.  If either value is less than the
    typical value (as determined by load_all_fonts() with a test string), the
    typical value is returned instead.  The line extends up to the first
    newline or the end of the string.  The string is assumed to be in UTF-8
    encoding.

    \param line Pointer to text position and style
    \param height Pointer to write height result to
    \param depth Pointer to write depth result to
    \param next Pointer to struct for text position and style of next line
*/
void line_height(const text_pos *line, int *height, int *depth, text_pos *next)
{
    const otb_glyph *gl;
    const char *scan;
    int style, maxheight, maxdepth;

    scan= line->pos;
    style= line->style;
    maxheight= typglypheight;
    maxdepth= typglyphdepth;
    while( *scan && *scan != '\n' )
    {
        if( *(const uchar *)scan < CHAR_PRINT_MIN ) { /* ignore control chars */
            if( isstyle(*scan) )
                style= *scan;
            ++scan;
            continue;
        }
        gl= get_glyph(utf8_codepoint(scan, &scan), style);
        if( gl->metrics->BearingY > maxheight )
            maxheight= gl->metrics->BearingY;
        if( gl->metrics->height - gl->metrics->BearingY > maxdepth )
            maxdepth= gl->metrics->height - gl->metrics->BearingY;
    }
    *height= maxheight;
    *depth= maxdepth;
    if( next ) {
        next->pos= (char*)(*scan ? scan+1 : scan);
        next->style= style;
    }
}


/*!
    \brief Paragraph skip for the current font set

    \return Number of pixels to advance the write position in y direction
    between paragraphs
*/
int parskip(void)
{
    return parskip_;
}



