/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#include <string.h>

#include "serialreader.h"


/*------------------ Line breaking functions ------------------------*/

typedef struct {
    text_pos currend, prevend;
    int width;
    autoarray *linelist;
}
break_state;

int advance_word(break_state *st);
int mark_eol(break_state *st);
int string_width(const text_pos *start, text_pos *end);
int break_lines(break_state *st, int maxwidth);

/* return codes for advance_word() and mark_eol() */
#define LBREAK_OK       0   /* continue processing */
#define LBREAK_ERROR    1   /* abort with error */
#define LBREAK_DONE     2   /* end of text encountered */
#define LBREAK_RESTART  3   /* for advance() to indicated forced line break */


/*!
    \brief Advance to first space after next word

    Advances the current position and style in the break_state struct and
    computes the new rendered line width using string_width().

    \param st Pointer to break_state struct containing the current position and
    style

    \return LBREAK_RESTART if a newline character is found, LBREAK_DONE if the
    end of the string is encountered, otherwise LBREAK_OK
*/
int advance_word(break_state *st)
{
    char *scan;
    int atlinestart= st->currend.pos == st->prevend.pos;

    scan= st->currend.pos;
    while( *scan && *scan <= ' ' && *scan != CHAR_UNKNOWN && *scan != '\n' )
        ++scan;
    if( *scan == '\n' ) {
        st->prevend= st->currend;
        st->currend.pos= scan;
        string_width(&st->prevend, &st->currend);  /* update style */
        return LBREAK_RESTART;
    }
    while( *scan && *scan != ' ' && *scan != '\n' )
        ++scan;         /* need not check for UTF-8 characters */
    st->prevend= st->currend;
    st->currend.pos= scan;
    st->width += string_width(&st->prevend, &st->currend);
    return *scan ? LBREAK_OK : LBREAK_DONE;
}


/*!
    \brief Width of a string when typeset with a given font

    Adds up the Advance values of all characters in a UTF-8 encoded string, up
    to the end pointer or the end of the string.  Control characters (< 32)
    are ignored except for style control bytes.

    \param start Pointer to start position and style
    \param end  Pointer to end position and style (style will be updated)

    \return String width
*/
int string_width(const text_pos *start, text_pos *end)
{
    const otb_glyph *gl;
    const char *str;
    int style, width;

    style= start->style;
    width= 0;
    str= start->pos;
    while( *str && str < end->pos )
    {
        if( *(const uchar *)str < CHAR_PRINT_MIN ) {
            if( isstyle(*str) )
                style= *str;
            ++str;
            continue;
        }
        gl= get_glyph(utf8_codepoint(str, &str), style);
        width += gl->advance;
    }
    end->style= style;
    return width;
}


/*!
    \brief Write a linefeed character and erase spaces

    A linefeed character is written to the current position, which must be at
    the first white-space character between words.  Other spaces and newlines
    in this inter-word gap are replaced by CHAR_ERASED which indicates they
    should not be typeset and have no horizontal advance.  Both the current and
    the previous text position are placed at the first following printable
    character or CHAR_STOP.  If the linefeed position points to a NUL
    character, no action is taken, and LBREAK_DONE is returned.

    \param st Pointer to break_state struct containing position and style

    \return LBREAK_DONE if the line break position is at a NUL character,
    otherwise LBREAK_OK
*/
int mark_eol(break_state *st)
{
    text_pos linestart;
    char *scan, *skip;

    scan= st->currend.pos;
    if( !*scan )
        return LBREAK_DONE;
    *scan++ = '\n';
    linestart.pos= scan;
    linestart.style= st->currend.style;
    autoappend(st->linelist, &linestart);
    if( *scan == CHAR_PARAGRAPH )
        ++scan;
    /* skip spaces after line break */
    for( skip= scan; utf8_isspace(skip, (const char **)&skip) ||
            ((*skip == CHAR_ERASED || *skip == CHAR_PARAGRAPH) && ++skip); );
    if( skip > scan )
        memset(scan, CHAR_ERASED, skip - scan);
    st->prevend= st->currend= linestart;
    st->width= 0;
    return LBREAK_OK;
}


/*!
    \brief Break text into lines

    Executes the line break algorithm by repeatedly calling advance_word() (to
    lengthen a line experimentally and compute the rendered width) and
    mark_eol() to mark the line break position and add the next line to the
    line array.  Lines are broken at white space between words unless a single
    word is already too long.

    \param st Pointer to struct containing current position, style and line
    width
    \param maxwidth Maximum allowed width of a line, will only be exceeded if a
                    single "word" is already longer.

    \return 0 on success, 1 if one of the auxiliary functions returned
    LBREAK_ERROR
*/
int break_lines(break_state *st, int maxwidth)
{
    int astatus, mstatus;

    do
    {
        astatus= advance_word(st);
        if( astatus == LBREAK_OK ) {
            do {
                astatus= advance_word(st);
            }
            while( astatus == LBREAK_OK && st->width <= maxwidth );
            if( st->width > maxwidth )
                st->currend= st->prevend;
        }
        if( astatus == LBREAK_ERROR )
            return 1;
        mstatus= mark_eol(st);
    }
    while( astatus != LBREAK_DONE && mstatus == LBREAK_OK );

    return astatus == LBREAK_ERROR || mstatus == LBREAK_ERROR;
}


/*!
    \brief Break document into lines

    Calls break_lines() after preparing the break_state struct that is its
    argument.

    \param doc Pointer to document text, concluded by NUL
    \param linelist Pointer to autoextensible array for line start positions
    and styles
    \param margin Width of page margin in pixels
*/
void break_document(char *doc, autoarray *linelist)
{
    break_state st= { doc, STYLE_REGULAR, doc, STYLE_REGULAR, 0, linelist };
    autoarray first_line= { doc, STYLE_REGULAR };

    autoappend(linelist, &first_line);
    break_lines(&st, SCREEN_WIDTH - 2*config.margin);
}


