/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

/*
    serialreader main program

    Compile with: grep ^arm main.c | sh
arm-linux-gcc -o serialreader main.c config.c document.c navigation.c unicode.c encodings.c font.c render.c linebreak.c basics.c ../iliad_api.c ../iliad_app.c -lpthread -leripc -lerdm -lgtk-x11-2.0 -lgthread-2.0
*/

#include <string.h>
#include <unistd.h>
#include <signal.h>

#include <gtk/gtk.h>

#include <liberdm/display.h>
#include <liberipc/eripctoolbar.h>

#include "../iliad_api.h"
#include "../iliad_app.h"

#include "serialreader.h"

unsigned char *screen;

static int icons[]= { iconID_fontlarger, iconID_fontsmaller, iconID_back,
                      iconID_forward, -1 };

void wwi(void);
int update_icons(void);
void sighandler(int sig);

/*!
    \brief Main program

    Initialises various parts of the application, then calls the event loop
    wwi().  When that function returns, cleans up and returns.

    \param argc Number of command line arguments (including own command name)
    \param argv List of command line arguments.  The first (the command name)
    is passed to GTK via app_init().  The second, the file to display, is
    passed to to_docdir() which makes it the default starting document
    (depending on file type and config file override).

    \return 0 on normal termination, 1 to 8 on failure at different steps in
    the initialisation
*/
int main(int argc, char **argv)
{
    GtkWidget *wind;
    GdkRectangle screenarea= { 0,0,SCREEN_WIDTH-1,CLIENT_AREA-1 };
    struct sigaction handler_arg;

    if( app_init(argv[0]) ) {
        fprintf(stddiag(), "app_init() failed.\n");
        return 1;
    }

    if( argc < 2 ) {
        fprintf(stddiag(), "Missing file or folder command-line argument.\n");
        return 2;
    }
    if( to_docdir(argv[1]) ) {
        fprintf(stddiag(), "Could not cd to document directory.\n");
        return 3;
    }
    if( read_config() ) {
        fprintf(stddiag(), "Could not read config file(s).\n");
        return 4;
    }
    if( load_all_fonts() ) {
        fprintf(stddiag(), "Could not load fonts.\n");
        return 5;
    }
    screen= framebuffer();
    if( !screen ) {
        fprintf(stddiag(), "Could not mmap framebuffer.\n");
        return 6;
    }
    if( change_file(NAV_INIT) ) {
        fprintf(stddiag(), "Could not load initial document.\n");
        return 7;
    }
    if( !can_change_fontsize(NAV_NEXT) )
        icons[0] |= (iconState_grey << 8);
    if( !can_change_fontsize(NAV_PREV) )
        icons[1] |= (iconState_grey << 8);
    if( !have_history(NAV_PREV) )
        icons[2] |= (iconState_grey << 8);
    if( !have_history(NAV_NEXT) )
        icons[3] |= (iconState_grey << 8);
    if( init_icons(icons) ) {
        fprintf(stddiag(), "Could not initialise icons.\n");
        return 8;
    }
    update_icons();     /* initialisation */
    if( init_pagebar(numdocs(), currdoc() + 1, -1, currdoc() > numdocs()-10) ) {
        fprintf(stddiag(), "Could not initialise page bar.\n");
        return 8;
    }

    memset(&handler_arg, 0, sizeof(handler_arg));
    handler_arg.sa_handler= sighandler;
    sigaction(SIGTERM, &handler_arg, NULL);
    sigaction(SIGABRT, &handler_arg, NULL);

    update_now4bit();
    set_led(LED_OFF);

    wwi();

    write_config();
    app_exit();
    return 0;
}


/*!
    \brief Wetware interface

    Event loop reading events from the joint event queue of the iliad_app
    interface.  Takes appropriate action on button presses, messages from
    pagebar and toolbar, and pen clicks.
*/
void wwi(void)
{
    iliad_event ev;
    int nav_dir, nav_file;
    int pagebarpage= currdoc();

    while( 13 )
    {
        app_wait_event(&ev);
//        fprintf(app_debug(), "event type %d subtype %d\n", ev.type, ev.sys.subtype);
        switch( ev.type ) {
            case EV_KEY:
                if( ev.key.key == EV_KEY_ASCEND )
                    return;
                else if( ev.key.key == EV_KEY_DOT ) {
                    if( ev.key.long_press )
                        nav_dir= have_history(NAV_NEXT) ? NAV_LAST : NAV_FIRST;
                    else
                        nav_dir= have_history(NAV_NEXT) ? NAV_NEXT : NAV_PREV;
                    nav_history(nav_dir);
                }
                else {
                    nav_file= 0;
                    switch(ev.key.key) {
                        case EV_KEY_DOWN:nav_file= 1;
                        case EV_KEY_FORW:
                            nav_dir= ev.key.long_press? NAV_LAST : NAV_NEXT;
                            break;
                        case EV_KEY_UP:nav_file= 1;
                        case EV_KEY_BACK:
                            nav_dir= ev.key.long_press? NAV_FIRST : NAV_PREV;
                            break;
                        default:continue;
                    }
                    if( nav_file )
                        change_file(nav_dir);
                    else
                        change_page(nav_dir);
                }
                break;
            case EV_PAGEBAR:
                if( ev.pagebar.page == currdoc() || ev.pagebar.page <= 0 )
                    continue;
                change_file(NAV_SET + ev.pagebar.page - 1);
                break;
            case EV_PEN:
                if( !near_posbar(ev.pen.xto, ev.pen.yto) )
                    continue;
                posbar_page(ev.pen.xto);
                break;
            case EV_TOOLBAR:
                if( ev.toolbar.state == iconState_grey )
                    continue;
                if( ev.toolbar.icon == iconID_fontlarger ||
                        ev.toolbar.icon == iconID_fontsmaller )
                    change_fontsize(ev.toolbar.icon == iconID_fontlarger?
                                    NAV_NEXT : NAV_PREV);
                else if( ev.toolbar.icon == iconID_back ||
                         ev.toolbar.icon == iconID_forward )
                    nav_history(ev.toolbar.icon == iconID_back ? NAV_PREV : NAV_NEXT);
                break;
            default:continue;
        }
        if( pagebarpage != currdoc() ) {
            set_pagebar_page(currdoc() + 1, -1);
            pagebarpage= currdoc();
        }
        update_icons();
        update_now4bit();
        if( ev.type == EV_KEY )
            set_led(LED_OFF);
    }
}


/*!
    \brief Update icon (toolbar) state

    This function determines the appropriate icon state from the state of other
    parts of the program, such as whether smaller and larger fonts are
    available.  On the first call, this state is only saved (under the
    assumption that the icons are correct).  On subsequent calls, the states of
    icons is updated as needed.

    \return 1 if an icon was changed, 0 otherwise
*/
int update_icons(void)
{
    static int icon_ids[4] = { iconID_fontlarger, iconID_fontsmaller,
                               iconID_back, iconID_forward },
               icon_states[4] = { -1 };

    int new_states[4];
    int count, change;

    new_states[0]= can_change_fontsize(NAV_NEXT) ? 0 : iconState_grey;
    new_states[1]= can_change_fontsize(NAV_PREV) ? 0 : iconState_grey;
    new_states[2]= have_history(NAV_PREV) ? 0 : iconState_grey;
    new_states[3]= have_history(NAV_NEXT) ? 0 : iconState_grey;
    change= 0;
    if( icon_states[0] < 0 ) {
        icon_states[0]= new_states[0];
        icon_states[1]= new_states[1];
        icon_states[2]= new_states[2];
        icon_states[3]= new_states[3];
    }
    else {
        tbDisableUpdate(toolbar_channel(), UAID);
        for( count= 0; count< 4; ++count )
            if( new_states[count] != icon_states[count] ) {
                tbSetStatePlatformIcon(toolbar_channel(), UAID,
                                       icon_ids[count], new_states[count]);
                icon_states[count]= new_states[count];
                change= 1;
            }
        tbEnableUpdate(toolbar_channel(), UAID);
    }
    return change;
}


/*!
    \brief TERM signal handler

    This function is installed as a signal handler for the TERM and ABRT
    signals by the main program.  It saves the current file and position using
    write_config() and terminates the program orderly via app_exit().

    \return This function never returns, but terminates the application
*/
void sighandler(int sig)
{
    write_config();
    app_exit();
    _exit(0);
}


