/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

#include "serialreader.h"

#define NAV_NO_HISTORY  0x40000000

typedef struct {
    int file, line, offset;
}
history_entry;

typedef struct {
    int line, offset;
}
last_pos;

sr_document *document;

static autoarray docfiles= { NULL, 0, 0, NAME_MAX+1, 0, 16 };
static last_pos *lastpositions;
static int currfileind= -1;

static autoarray history= { NULL, 0, 0, sizeof(history_entry), 0, 16 };
static int hist_ind= 0;

#define filenamecpy(dest, entry)    \
                { strncpy((dest), (entry)->d_name, NAME_MAX); \
                (dest)[NAME_MAX]= 0; }

#define filenamecmp(name, entry)  strcasecmp((name), (entry)->d_name)

int list_files();
int txtorhtml(const struct dirent *entry);
void history_push(void);
void history_update(void);
void history_fake(int fileind, int line, int offset);
int history_go(int newhistind);
int closest_line(int offset);
int cmplinepos(const void *target, const void *line);


/*!
    \brief Change current directory to document directory and list files

    The document directory is saved to the global constants struct
    config, and the start file name is set to argv1 without the directory path.
    Then the current directory is changed and list_files() is called.

    \param argv1 Command-line argument giving file to display or directory of
    files to display

    \return 0 OK, otherwise chdir failed
*/
int to_docdir(char *argv1)
{
    struct stat st;
    char *scan, *filepart;
    int length, ind;

    if( stat(argv1, &st) )
        return 1;
    if( S_ISDIR(st.st_mode) )
        length= strlen(argv1);
    else {
        length= 0;
        for( ind= 0; argv1[ind]; ++ind )
            if( argv1[ind] == '/' )
                length= ind;
    }
    if( length ) {
        config.argv_file= argv1[length] ? argv1 + length + 1 : "";
        while( argv1[length-1] == '/' && length > 1 )
            --length;
        argv1[length]= 0;
        config.docdir= argv1;
    }
    else {
        config.argv_file= argv1;
        config.docdir= "";
    }
    return chdir(config.docdir) || list_files();
}


/*!
    \brief Create list of document files in the current directory.

    Reads entries of the current directory, test them with txtorhtml() and adds
    them to the docfiles array if accepted.  Space for the table of last
    positions in every file is allocated.

    \return 0 OK, 1 could not open current directory or could not allocate
    space for last position.
*/
int list_files()
{
    DIR *dir;
    struct dirent *entry;

    dir= opendir(".");
    if( !dir )
        return 1;
    while( entry= readdir(dir) )
        if( txtorhtml(entry) )
            autoappend(&docfiles, entry->d_name);
    closedir(dir);
    qsort(docfiles.buf, docfiles.nelem, docfiles.elemsize, strcase2cmp);
    lastpositions= malloc(sizeof(last_pos) * docfiles.nelem);
    if( !lastpositions )
        return 2;
    memset(lastpositions, 0, sizeof(last_pos) * docfiles.nelem);
    return 0;
}


/*!
    \brief Return number of document files available

    Must not be called before list_files().

    \return Number of files
*/
int numdocs(void)
{
    return (int)docfiles.nelem;
}


/*!
    \brief Return index of current document file

    Must not be called before change_file(NAV_INIT).

    \return Index of current document, 0-based
*/
int currdoc(void)
{
    return currfileind;
}


/*!
    \brief Decide if a directory entry refers to a plain text or HTML file

    Obviously this is guesswork.  File extensions .html and .htm (case
    insensitive) count as HTML, and .txt and files without extension count as
    plain text.  Files with the extension .ser are also accepted; anything but
    regular files is rejected.

    \param entry Pointer to directory entry struct

    \return 1 if text or HTML, 0 otherwise
*/
int txtorhtml(const struct dirent *entry)
{
    struct stat st;
    const char *scan, *dot;

    if( stat(entry->d_name, &st) || !S_ISREG(st.st_mode) )
        return 0;
    dot= NULL;
    for( scan= entry->d_name; *scan; ++scan )
        if( *scan == '.' )
            dot= scan;
    return !dot || !strcasecmp(dot+1, "html") || !strcasecmp(dot+1, "htm")
                || !strcasecmp(dot+1, "txt") || !strcasecmp(dot+1, "ser");
}


/*!
    \brief Change current document file

    Change to next, previous, first or last file; or initialise and load start
    file.

    \param mode NAV_INIT, NAV_FIRST, NAV_PREV, NAV_NEXT, NAV_LAST; or NAV_SET +
    the file index

    \return 0 OK, !=0 error.  In case of an error, the global document pointer
    may still point to the previous document struct, unless an error occurred
    while loading the new document.
*/
int change_file(int mode)
{
    char *newfile, *startfile;
    int no_history, newfileind, startfileind, newline, newoffset;

    no_history= !!(mode & NAV_NO_HISTORY);
    mode &= ~NAV_NO_HISTORY;
    if( mode == NAV_INIT ) {
        if( config.argv_file && *config.argv_file ) {
            newfile= bsearch(config.argv_file, docfiles.buf,
                             docfiles.nelem, docfiles.elemsize, strcase2cmp);
            newfileind= !newfile ? -1 : 
                        (newfile - (char*)docfiles.buf) / docfiles.elemsize;
        }
        else
            newfileind= -1;
        if( config.start_file && *config.start_file ) {
            startfile= bsearch(config.start_file, docfiles.buf,
                             docfiles.nelem, docfiles.elemsize, strcase2cmp);
            startfileind= !startfile ? -1 : 
                         (startfile - (char*)docfiles.buf) / docfiles.elemsize;
        }
        else
            startfileind= -1;
        if( newfileind >= 0 ) {
            newline= 0;
            newoffset= -1;
            if( startfileind >= 0 )
                history_fake(startfileind, -1, config.start_position);
        }
        else if( startfileind >= 0 ) {
            newfile= startfile;
            newfileind= startfileind;
            newline= -1;
            newoffset= config.start_position;
        }
        else {
            newfile= (char*)docfiles.buf;
            newfileind= 0;
            newline= 0;
            newoffset= -1;
        }
    }
    else  {
        switch( mode ) {
            case NAV_FIRST:newfileind= 0;
                        break;
            case NAV_LAST:newfileind= docfiles.nelem - 1;
                        break;
            case NAV_NEXT:newfileind= currfileind + 1;
                          if( newfileind >= docfiles.nelem )
                              newfileind= docfiles.nelem - 1;
                        break;
            case NAV_PREV:newfileind= currfileind - 1;
                          if( newfileind < 0 )
                              newfileind= 0;
                        break;
            default:newfileind= mode - NAV_SET;
                    if( newfileind >= docfiles.nelem )
                        newfileind= docfiles.nelem - 1;
                    break;
        }
        newfile= (char*)docfiles.buf + newfileind * docfiles.elemsize;
        newline= lastpositions[newfileind].line;
        newoffset= lastpositions[newfileind].offset;
    }
    if( document ) {
        if( newfileind == currfileind )
            return 0;
        lastpositions[currfileind].line= document->curr_first_line;
        lastpositions[currfileind].offset=
            ((text_pos*)document->lines.buf)[document->curr_first_line].pos
            - document->text;
        if( !no_history )
            history_update();
        discard_document(document);
    }
    currfileind= newfileind;
    document= load_document(newfile);
    if( !document )
        return 1;
    document->filename= newfile;
    if( newline >= 0 && newline < document->lines.nelem && (newoffset < 0 ||
                        ((text_pos*)document->lines.buf)[newline].pos
                        == document->text + newoffset) )
        document->curr_first_line= newline;
    else
        document->curr_first_line= closest_line(newoffset);
    if( !no_history )
        history_push();
    typeset_page(document, 0);
    return 0;
}


/*!
    \brief Go to next/previous/first/last page in the current document

    This function just calls typeset_page() with the approriate arguments.  The
    dirty work of fitting lines on the page (backwards) is done by fill_page().

    \param mode NAV_FIRST, NAV_PREV, NAV_NEXT, NAV_LAST; or NAV_SET + the line
    index
*/
void change_page(int mode)
{
    int step;

    if( mode == NAV_FIRST || mode == NAV_LAST )
        history_update();
    switch( mode ) {
        case NAV_FIRST:
        case NAV_INIT:document->curr_first_line= 0;
                      step= 0;
                    break;
        case NAV_LAST:document->curr_first_line= document->lines.nelem;
                      step= -1;
                    break;
        case NAV_NEXT:step= 1;
                    break;
        case NAV_PREV:step= -1;
                    break;
        default:document->curr_first_line= mode - NAV_SET;
                if( document->curr_first_line >= document->lines.nelem )
                    document->curr_first_line= document->lines.nelem - 1;
                step= 0;
                break;
    }
    typeset_page(document, step);
    if( mode == NAV_FIRST || mode == NAV_LAST )
        history_push();
}


/*!
    \brief Go to the position indicated by a click on the position bar.

    \param x Screen position in the position bar
*/
void posbar_page(int x)
{
    int lineind;

    lineind= read_posbar(document, x);
    history_update();
    change_page(lineind >= 0? NAV_SET + lineind : NAV_LAST);
    history_push();
}


/*!
    \brief Add current position as new history location

    Any existing forward history is erased by resetting the current history
    index.  After appending the current position, the index is incremented to
    make it the new current history position.
*/
void history_push(void)
{
    history_entry newhist=
            { currfileind, document->curr_first_line,
              ((text_pos*)document->lines.buf)[document->curr_first_line].pos
              - document->text };

    if( hist_ind + 1 < (int)history.nelem )
        history.nelem= hist_ind + 1;
    autoappend(&history, &newhist);
    hist_ind= history.nelem - 1;
}


/*!
    \brief Update current history location

    This is used to keep the line position up to date while visiting locations
    from the history.
*/
void history_update(void)
{
    ((history_entry*)history.buf)[hist_ind].file= currfileind;
    ((history_entry*)history.buf)[hist_ind].line= document->curr_first_line;
    ((history_entry*)history.buf)[hist_ind].offset=
            ((text_pos*)document->lines.buf)[document->curr_first_line].pos
            - document->text;
}


/*!
    \brief Add given position to history

    Any existing forward history is erased by resetting the current history
    index.  After appending the current position, the index is incremented to
    make it the new current history position.

    \param fileind File index
    \param line Line
    \param offset Byte offset from document start
*/
void history_fake(int fileind, int line, int offset)
{
    history_entry newhist= { fileind, line, offset };

    if( hist_ind + 1 < (int)history.nelem )
        history.nelem= hist_ind + 1;
    autoappend(&history, &newhist);
    hist_ind= history.nelem - 1;
}


/*!
    \brief Return whether it is possible to go to a different history position.

    \param mode NAV_FIRST, NAV_PREV, NAV_NEXT or NAV_LAST

    \return !=0: can go in the specified direction in history, 0 can't
*/
int have_history(int mode)
{
    if( mode == NAV_NEXT || mode == NAV_LAST )
        return hist_ind < (int)history.nelem - 1;
    else if( mode == NAV_PREV || mode == NAV_FIRST )
        return hist_ind;
    else
        return 0;
}


/*!
    \brief Go to a different history position

    If there is no more history available for NAV_NEXT or NAV_PREV, the
    operation will fail silently without reloading the current document.

    \param mode NAV_FIRST, NAV_PREV, NAV_NEXT or NAV_LAST

    \return 0 done, 1 error loading document or illegal mode
*/
int nav_history(int mode)
{
    int newhistind;

    switch( mode ) {
        case NAV_FIRST:newhistind= 0;
                break;
        case NAV_PREV:newhistind= hist_ind - 1;
                break;
        case NAV_NEXT:newhistind= hist_ind + 1;
                break;
        case NAV_LAST:newhistind= (int)history.nelem - 1;
                break;
        default:return 1;
    }
    if( newhistind < 0 || newhistind >= (int)history.nelem || newhistind == hist_ind )
        return 0;
    history_update();
    return history_go(newhistind);
}


/*!
    \brief Go to a position saved in the history

    Calls change_page() or change_file() as appropriate.

    \param newhistind Index of history entry to go to

    \return 0 done, 1 error loading document
*/
int history_go(int newhistind)
{
    history_entry *hist= ((history_entry*)history.buf) + newhistind;

    hist_ind= newhistind;
    if( hist->file == currfileind ) {
        if( hist->line < 0 || hist->line >= document->lines.nelem ||
                        ((text_pos*)document->lines.buf)[hist->line].pos
                        != document->text + hist->offset ) {
            hist->line= closest_line(hist->offset);
            hist->offset= ((text_pos*)document->lines.buf)[hist->line].pos -
                                    document->text;
        }
        change_page(NAV_SET + hist->line);
        return 0;
    }
    else {
        lastpositions[hist->file].line= hist->line;
        lastpositions[hist->file].offset= hist->offset;
        return change_file(NAV_NO_HISTORY | (NAV_SET + hist->file));
    }
}


static const text_pos *closest_elem;
static int closest_diff;

/*!
    \brief Find line starting closest to a text position

    \param offset Index of position in (pre-processed) document text

    \return Index of closest line in document's line array
*/
int closest_line(int offset)
{
    closest_diff= INT_MAX;
    bsearch(document->text+offset, document->lines.buf, document->lines.nelem,
            document->lines.elemsize, cmplinepos);
    return closest_elem - (const text_pos*)document->lines.buf;
}


/*!
    \brief Line start comparison function which determines closest match

    This function is used by closest_line in its call to bsearch.  It compares
    its first argument (a pointer into the document text) with the start of the
    line given by the second (a pointer to a text_pos struct in the line array
    of the document).  In addition to returning the appropriate value for
    bsearch(), it updates closest_elem to the line being compared if its
    distance from the target position is less than closest_diff.

    \param target Pointer to target position in the document text
    \param line Pointer to text_pos struct of a line

    \return Difference of the position pointers as required by bsearch()
*/
int cmplinepos(const void *target, const void *line)
{
    int diff= (char *)target - ((text_pos *)line)->pos;

    if( abs(diff) < closest_diff ) {
        closest_elem= line;
        closest_diff= abs(diff);
    }
    return diff;
}


/*!
    \brief Return whether the font size can be changed in the given direction

    \param mode NAV_FIRST, NAV_PREV, NAV_NEXT or NAV_LAST

    \return !=0: can change font size in the given direction
*/
int can_change_fontsize(int mode)
{
    if( mode == NAV_NEXT || mode == NAV_LAST )
        return config.fontsize < config.numfontsizes - 1;
    else if( mode == NAV_PREV || mode == NAV_FIRST )
        return config.fontsize;
    else
        return 0;
}


/*!
    \brief Change font size

    Changes the font size according to its argument, loads the new fonts,
    reloads the current document and displays a page at the current position.

    \param mode NAV_FIRST, NAV_PREV, NAV_NEXT or NAV_LAST

    \return 0 OK, 1 Font was not changed (because no smaller/larger font was
    available, or because of an error loading the new fonts), -1 error loading
    font (no font available) or document
*/
int change_fontsize(int mode)
{
    char *filename;
    int newfontsize, status;

    switch(mode) {
        case NAV_FIRST:newfontsize= 0;
                break;
        case NAV_PREV:newfontsize= config.fontsize - 1;
                break;
        case NAV_NEXT:newfontsize= config.fontsize + 1;
                break;
        case NAV_LAST:newfontsize= config.numfontsizes - 1;
                break;
        default:return 1;
    }
    if( newfontsize == config.fontsize || newfontsize < 0 || newfontsize >= config.numfontsizes )
        return 1;
    config.fontsize= newfontsize;
    status= load_all_fonts();
    if( status )
        return status;
    history_update();
    filename= document->filename;
    discard_document(document);
    document= load_document(filename);
    if( !document )
        return -1;
    document->filename= filename;
    history_go(hist_ind);
    return 0;
}

