/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#include <string.h>

#include "serialreader.h"

#define MAX_TEXT_LINES  (CLIENT_AREA / 10)

typedef struct {
    char *pos;
    int style;
    uchar *draw;
    int lmargin, y;
}
render_state;

/* Minimum text y position; and coordinates of the position bar without frame */
static int text_min_y, posbar_x0, posbar_y0, posbar_x1, posbar_y1;
static int line_ypos[MAX_TEXT_LINES+1];

void typeset_line(render_state *st);
void fill_page(sr_document *doc, int pagestep);
void render_headline(const sr_document *doc);
void render_posbar(const sr_document *doc);

/*!
    \brief Typeset one line of text

    Typesets a line of text and updates the current position and style in the
    render_state struct.  Clipping in the y direction is performed so that
    partially visible lines can be typeset, but partially visible characters at
    the end of the line are discarded.

    \param st Pointer to render state struct containing text position, drawing
    position, style and margin values.  The position after this function
    returns will be on the newline ending the line.  The style will have been
    updated.
*/
void typeset_line(render_state *st)
{
    const otb_glyph *gl;
    unsigned char *read, *img, *glyphpos, *write, bits;
    int shift, x, y, ytop;

    glyphpos= st->draw + st->lmargin;
    while( *st->pos && *st->pos != '\n' && glyphpos < st->draw + SCREEN_WIDTH )
    {
        if( *(unsigned char *)st->pos < CHAR_PRINT_MIN ) {
            if( isstyle(*st->pos) )
                st->style= *st->pos;
            ++st->pos;
            continue;
        }
        gl= get_glyph(utf8_codepoint(st->pos, (const char**)&st->pos), st->style);
        shift= 0;
        read= (unsigned char *)gl->bitmap;
        bits= *read;
        write= glyphpos - gl->metrics->BearingY * SCREEN_WIDTH + gl->metrics->BearingX;
        if( (int)gl->metrics->width > SCREEN_WIDTH - (glyphpos - st->draw) )
            break;
        ytop= gl->metrics->height;
        if( st->y - gl->metrics->BearingY + ytop > CLIENT_AREA )
            ytop= CLIENT_AREA - st->y + gl->metrics->BearingY;
        for( y= 0; y< ytop; ++y ) {
            for( x= 0; x< gl->metrics->width; ++x ) {
                if( shift >= 8 ) {
                    shift= 0;
                    ++read;
                    bits= *read;
                }
                if( bits & 0xC0 )
                    write[x] = 255 - (bits & 0xC0) / 0x40 * 85;
                bits <<= 2;
                shift += 2;
            }
            if( !gl->bitpacked ) {
                shift= 0;
                ++read;
                bits= *read;
            }
            write += SCREEN_WIDTH;
        }
        glyphpos += gl->advance;
    }
}


/*!
    \brief Determine y positions of text lines on the screen

    First the index of the first visible line is computed if the pagestep
    parameter indicates scrolling up or down.  For going back by one page, the
    number of prior lines which fit on the page has to be computed.  The
    current first visible line in the document struct is updated accordingly.

    Starting from the first line to be displayed, the height and depth of each
    line are determined using line_height(), which takes unusually high glyphs
    into account.  The vertical pixel positions are written to the array
    line_ypos, concluded by -1.  The index of the last visible line is written
    to the element curr_last_line of the document struct, unless the end of the
    document was reached, when it is set to -1.

    \param doc Pointer to document data struct.  The current visible line range
    will be updated.
    \param pagestep Signed flag indicating whether the current, previous (<0)
    or next page (>0) should be displayed
*/
void fill_page(sr_document *doc, int pagestep)
{
    text_pos *lines;
    int *pos;
    int nlines, linenr, ypos, height, depth;

    lines= (text_pos*)doc->lines.buf;
    nlines= doc->lines.nelem;
    if( pagestep < 0 ) {
        linenr= doc->curr_first_line + 1;
        if( linenr >= nlines )
            linenr= nlines - 1;
        line_height(lines+linenr, &height, &depth, NULL);
        ypos= 2 + (height + 5) / 10 + height;
        if( linenr == nlines-1 )
            ypos += 1 + (depth + 5) / 10 + depth;
        if( *lines[linenr].pos == CHAR_PARAGRAPH )
            ypos += parskip();
        for( --linenr; linenr >= 0; --linenr ) {
            line_height(lines+linenr, &height, &depth, NULL);
            ypos += 2 + (height + 5) / 10 + height;
            ypos += 1 + (depth + 5) / 10 + depth;
            if( ypos > CLIENT_AREA - text_min_y )
                break;
            if( *lines[linenr].pos == CHAR_PARAGRAPH )
                ypos += parskip();
        }
        doc->curr_first_line= linenr + 1;
    }
    else if( pagestep > 0 && doc->curr_last_line >= 0 ) {
            linenr= doc->curr_last_line;
            if( linenr > doc->curr_first_line + 5 )
                --linenr;           /* two lines overlap between pages */
            doc->curr_first_line= linenr;
        }
    linenr= doc->curr_first_line;
    if( linenr >= nlines )      /* may occur for stale .cfg file */
        linenr= nlines - 1;
    ypos= text_min_y;
    pos= line_ypos;
    doc->curr_last_line= -1;
    while( pos < line_ypos + MAX_TEXT_LINES && linenr < nlines ) {
        line_height(lines+linenr, &height, &depth, NULL);
        if( *lines[linenr].pos == CHAR_PARAGRAPH && pos > line_ypos )
            ypos += parskip();
        ypos += 2 + (height + 5) / 10 + height;
        if( ypos > CLIENT_AREA ) {
            doc->curr_last_line= linenr - 1;
            break;
        }
        *pos++ = ypos;
        ypos += 1 + (depth + 5) / 10 + depth;
        ++linenr;
    }
    *pos= -1;
}


/*!
    \brief Typeset a page of text

    Typesets a page of text line by line using typeset_line().  The vertical
    positions of lines are determined by fill_page(), after advancing or going
    back by one page if the pagestep parameter indicates that.  A partially
    visible last line will be typeset if most of it is visible.  The elements
    curr_first_line and curr_last_line of the document struct are updated by
    fill_page().

    \param doc Pointer to document data struct.  The current visible line range
    will be updated.
    \param pagestep Signed flag indicating whether the current, previous (<0)
    or next page (>0) should be displayed
    \param screen Pointer to the drawing area.  Its width and height is
    hard-coded.
*/
void typeset_page(sr_document *doc, int pagestep)
{
    render_state st;
    text_pos *linelist;
    int lineind;

    render_headline(doc);
    fill_page(doc, pagestep);
    st.lmargin= config.margin;
    st.y= text_min_y;
    memset(screen+text_min_y*SCREEN_WIDTH, 0xFF, (CLIENT_AREA-text_min_y)*SCREEN_WIDTH);
    linelist= (text_pos*)doc->lines.buf;
    for( lineind= 0; line_ypos[lineind] >= 0; ++lineind ) {
        st.pos= linelist[doc->curr_first_line+lineind].pos;
        st.style= linelist[doc->curr_first_line+lineind].style;
        st.y= line_ypos[lineind];
        st.draw= screen + st.y * SCREEN_WIDTH;
        typeset_line(&st);
    }
    render_posbar(doc);
}


#define FILENAMEWIDTH   (SCREEN_WIDTH/2)
#define POSBARVMARGIN    4
#define POSBARHMARGIN    15
#define RENDERFILENAME_MAX    256

/*!
    \brief Render file name and position bar frame in headline

    The file path is composed of the document directory and the current file
    name.  If it differs from the previous one, the file path, the separator
    line and the frame around the position bar are redrawn.

    \param doc Pointer to document data struct containing the current file name
*/
void render_headline(const sr_document *doc)
{
    static char rendered_filename[RENDERFILENAME_MAX+1];
// TODO: redraw when font changes ?

    char current_filename[RENDERFILENAME_MAX+1];
    render_state fname_render;
    const char *filenamestart;
    const unsigned char *read, *truncpos;
    unsigned char *drawline;
    const otb_glyph *gl;
    int dirlen, fnamelen, pathlen, glwidth, dotwidth, width, height, depth, drawind, posbarwidth, y;

    dirlen= strlen(config.docdir);
    fnamelen= strlen(doc->filename);
    pathlen= dirlen + 1 + fnamelen;
    if( fnamelen >= RENDERFILENAME_MAX )
        strcpy(current_filename, doc->filename + fnamelen - RENDERFILENAME_MAX);
    else if( !dirlen )
        strcpy(current_filename, doc->filename);
    else if( pathlen > RENDERFILENAME_MAX ) {
        strcpy(current_filename, config.docdir + pathlen - RENDERFILENAME_MAX);
        current_filename[RENDERFILENAME_MAX - fnamelen - 1]= '/';
        strcpy(current_filename + RENDERFILENAME_MAX - fnamelen, doc->filename);
    }
    else {
        strcpy(current_filename, config.docdir);
        current_filename[dirlen]= '/';
        strcpy(current_filename + dirlen + 1, doc->filename);
    }
    if( strcmp(rendered_filename, current_filename) ) {
        strcpy(rendered_filename, current_filename);
        gl= get_glyph('.', STYLE_REGULAR);
        dotwidth= gl->advance;
        /* Determine how much of the file name fits, starting at the back: */
        width= 0;
        read= rendered_filename + strlen(rendered_filename) - 1;
        while( width < FILENAMEWIDTH && (char*)read >= rendered_filename ) {
            gl= get_glyph(*read, STYLE_REGULAR);
            glwidth= gl->advance;
            width += glwidth;
            if( width < FILENAMEWIDTH - 3*dotwidth )
                truncpos= read;
            --read;
        }
        /* Set file name and text area position and typeset file name: */
        fname_render.pos= (char*)(width > FILENAMEWIDTH ? truncpos : read + 1);
        fname_render.style= STYLE_REGULAR;
        line_height((text_pos*)&fname_render, &height, &depth, NULL);
        text_min_y= height + depth + parskip() + 2;
        memset(screen, 0xFF, (text_min_y-2) * SCREEN_WIDTH);
        memset(screen + (text_min_y-2) * SCREEN_WIDTH, 0, 2 * SCREEN_WIDTH);
        fname_render.lmargin= 0;
        fname_render.y= parskip()/2 + height;
        fname_render.draw= screen + (fname_render.y + 1) * SCREEN_WIDTH - FILENAMEWIDTH;
        if( width > FILENAMEWIDTH ) {
            fname_render.pos= "...";
            typeset_line(&fname_render);
            fname_render.pos= (char*)truncpos;
            fname_render.draw += 3 * dotwidth;
        }
        typeset_line(&fname_render);
        /* Set position bar coordinates and draw frame: */
        posbar_x0= POSBARHMARGIN + 2;
        posbar_y0= POSBARVMARGIN + 2;
        posbar_x1= SCREEN_WIDTH - FILENAMEWIDTH - POSBARHMARGIN - 3;
        posbar_y1= text_min_y - 2 - POSBARVMARGIN - 3;
        posbarwidth= SCREEN_WIDTH - FILENAMEWIDTH - 2*POSBARHMARGIN;
        drawline= screen + (posbar_y0-2)*SCREEN_WIDTH + posbar_x0 - 2;
        for( y= posbar_y0 - 2; y < posbar_y0; ++y, drawline += SCREEN_WIDTH )
            for( drawind= 0; drawind < posbarwidth; ++drawind )
                drawline[drawind]= 0;
        for( ; y <= posbar_y1; ++y, drawline += SCREEN_WIDTH )
            drawline[0]= drawline[1]= drawline[posbarwidth-2]= drawline[posbarwidth-1]= 0;
        for( ; y <= posbar_y1 + 2; ++y, drawline += SCREEN_WIDTH )
            for( drawind= 0; drawind < posbarwidth; ++drawind )
                drawline[drawind]= 0;
    }
}


/*!
    \brief Redraw the inner part of the position bar

    The inner part of the position bar is redrawn to match the current position.

    \param doc Pointer to document data struct containing the current visible
    line range
*/
void render_posbar(const sr_document *doc)
{
    unsigned char *drawline;
    int posbarwidth, y, pos0, pos1, x;

    posbarwidth= posbar_x1 - posbar_x0 + 1;
    pos0= (doc->curr_first_line * posbarwidth) / doc->lines.nelem;
    if( doc->curr_last_line < 0 ) {
        pos1= posbarwidth - 1;
        if( pos0 > pos1 - 1 )
            pos0= pos1 - 1;
    }
    else {
        pos1= (doc->curr_last_line * posbarwidth) / doc->lines.nelem;
        if( pos1 - pos0 < 1 )
            if( pos0 >= posbarwidth - 2 ) {
                pos0= posbarwidth - 2;
                pos1= posbarwidth - 1;
            }
            else
                pos1= pos0 + 1;
    }
    drawline= screen + posbar_y0 * SCREEN_WIDTH + posbar_x0;
    for( y= posbar_y0; y <= posbar_y1; ++y ) {
        for( x= 0; x < pos0; ++x )
            drawline[x]= 0xFF;
        for( ; x <= pos1; ++x )
            drawline[x]= 0;
        for( ; x < posbarwidth; ++x )
            drawline[x]= 0xFF;
        drawline += SCREEN_WIDTH;
    }
}


/*!
    \brief Read file position off from screen coordinate in position bar

    \param x Screen coordinate in position bar

    \return -1: End of file (position to the right of posbar); otherwise line
    index
*/
int read_posbar(sr_document *doc, int x)
{
    if( x < posbar_x0 )
        return 0;
    else if( x > posbar_x1 )
        return -1;
    else
        return (x - posbar_x0) * doc->lines.nelem / (posbar_x1 - posbar_x0 + 1);
}


/*!
    \brief Decide if coordinates are near position bar

    This function decides if the position of a pen click is sufficiently close
    to the position bar to be interpreted as a new position click.

    \param x X coordinate
    \param y Y coordinate

    \return 1 == near position bar, 0 not
*/
int near_posbar(int x, int y)
{
    return x < SCREEN_WIDTH - FILENAMEWIDTH && y <= text_min_y;
}

