/*
    This file is part of serialreader, an HTML reader for the iRex Iliad,
    and is copyright (c) 2010 - 2012 Volker Schatz.

    It is licensed under Version 3 or later of the Gnu General Public License
    (see http://www.gnu.org/licenses/gpl.html for details).
*/

#include "serialreader.h"

/*!
    \brief Check for valid UTF-8

    Checks whether a character in a text is valid UTF-8.

    \param text Pointer to the putative UTF-8 character
    \param next Pointer to which the position of the next character is written
    if and only if the current character is valid UTF-8

    \return 1 if valid UTF-8, 0 otherwise
*/
int valid_utf8(const char *text, const char **next)
{
    unsigned c= *(const unsigned char*)text;
    unsigned morechars, ind;

    if( c < 0x80 ) {
        if( next )
            *next= text + 1;
        return 1;
    }
    if( c < 0xC0 || c > 0xF8 )
        return 0;
    if( c < 0xE0 )
        morechars= 1;
    else if( c < 0xF0 )
        morechars= 2;
    else
        morechars= 3;
    for( ind= 1; ind <= morechars; ++ind )
        if( (text[ind] & 0xC0) != 0x80 )
            return 0;
    if( next )
        *next= text + 1 + morechars;
    return 1;
}


/*!
    \brief Get code point of current UTF-8 character.

    The unicode code point of a character in UTF-8 encoded text is determined.
    The text is assumed to be correct UTF-8, and no error checking is done.
    If the second argument is not NULL, the pointer to the next character will
    be written to it, but not beyond a zero character (then it will be assigned
    the original text pointer).

    \param text Pointer to UTF-8 encoded character
    \param next Pointer to write pointer to next UTF-8 character to, or NULL

    \return Unicode code point
*/
int utf8_codepoint(const char *text, const char **next)
{
    int c= *(const signed char*)text;
    int mask;

    if( c >= 0 ) {
        if( next )
            *next= c ? text + 1 : text;
        return c;
    }
    ++text;
    mask= 0xFFFFFFC0;
    while( (*text & 0xC0) == 0x80 ) {
        c <<= 6;
        c |= *text & 0x3F;
        mask <<= 5;
        ++text;
    }
    c &= ~mask;
    if( next )
        *next= text;
    return c;
}


/*
    \brief Find out if a Unicode character is a space

    \param text Pointer to UTF-8 encoded character
    \param next Pointer to pointer which will be set to the position of the
            next character IF THERE WAS A SPACE, otherwise the current position
    \return 1 if space, 0 otherwise
*/
int utf8_isspace(const char *text, const char **next)
{
    unsigned codepoint= utf8_codepoint(text, next);

    if( isspace(codepoint) || codepoint == 0xA0 ||
            (codepoint >= 0x2000 && codepoint <= 0x200D) ||
            codepoint == 0x202F || codepoint == 0x205F || codepoint == 0xFEFF )
        return 1;
    if( next )
        *next= text;
    return 0;
}


/*!
    \brief Write unicode code point as UTF-8

    Writes a unicode character in UTF-8 encoding.  Up to four bytes may be
    written.  Out-of-range code points are replaced with the unknown character
    marker CHAR_UNKNOWN.

    \param codepoint Unicode code point
    \param dest Pointer to buffer with sufficient space
    \param end Pointer to pointer to new write position after this character,
    or NULL
*/
void utf8_write(int codepoint, char *dest, char **end)
{
    char continuation[3];
    int bytes, mask;

    if( (unsigned)codepoint < 0x80 )
        *dest++ = codepoint;
    else if( (unsigned)codepoint > 0x10FFFF )
        *dest++ = CHAR_UNKNOWN;
    else {
        bytes= 0;
        mask= 0x3F;
        do {
            continuation[bytes]= codepoint & 0x3F;
            ++bytes;
            codepoint >>= 6;
            mask >>= 1;
        }
        while( codepoint > mask );
        *dest++ = (codepoint & mask) | (0xF80 >> bytes);
        for( --bytes; bytes >= 0; --bytes )
            *dest++ = 0x80 | continuation[bytes];
    }
    if( end )
        *end= dest;
}


/*!
    \brief Skip a UTF-8 character.

    Skips a character in UTF-8 encoded text, but does not skip a zero byte.
    The text is assumed to be correct UTF-8, and no error checking is done.

    \param text Pointer to pointer to UTF-8 encoded text
*/
void utf8_skip(const char **text)
{
    if( *(const signed char *)*text >= 0 ) {
        if( **text )
            ++ *text;
    }
    else {
        do { ++ *text;
        }
        while( (**text & 0xC0) == 0x80 );
    }
}


/*!
    \brief Skip a UTF-8 character backwards.

    Skips back a character in UTF-8 encoded text.  The pointer then points to
    the first byte of the UTF-8 character (as it must before).
    The text is assumed to be correct UTF-8, and no error checking is done.

    \param text Pointer to pointer to UTF-8 encoded text
*/
void utf8_rskip(const char **text)
{
    do { -- *text;
    }
    while( (**text & 0xC0) == 0x80 );
}


